from decimal import Decimal

import pytest

from bo4e import Mengeneinheit, Preisstaffel, Preistyp, Tarifpreisposition, Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip


class TestTarifpreisposition:
    @pytest.mark.parametrize(
        "tarifpreisposition",
        [
            pytest.param(
                Tarifpreisposition(
                    preistyp=Preistyp.ENTGELT_ABLESUNG,
                    einheit=Waehrungseinheit.EUR,
                    bezugseinheit=Mengeneinheit.KWH,
                    preisstaffeln=[
                        Preisstaffel(
                            einheitspreis=Decimal(40.0),
                            staffelgrenze_von=Decimal(12.5),
                            staffelgrenze_bis=Decimal(25.0),
                        ),
                    ],
                    mengeneinheitstaffel=Mengeneinheit.STUECK,
                ),
                id="optional and required attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifpreisposition: Tarifpreisposition) -> None:
        """
        Test de-/serialisation of Tarifpreisposition.
        """
        assert_serialization_roundtrip(tarifpreisposition)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Kontaktart(StrEnum):
    """
    Gibt an, auf welchem Weg die Person oder der Geschäftspartner kontaktiert werden kann.
    """

    POSTWEG = "POSTWEG"  #: Postweg
    TELEFON = "TELEFON"  #: Telefon
    FAX = "FAX"  #: Fax
    E_MAIL = "E_MAIL"  #: E-Mail
    SMS = "SMS"  #: SMS



"""
Contains Ausschreibungslos class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import Optional

from ..enum.preismodell import Preismodell
from ..enum.rechnungslegung import Rechnungslegung
from ..enum.sparte import Sparte
from ..enum.vertragsform import Vertragsform
from ..utils import postprocess_docstring
from .ausschreibungsdetail import Ausschreibungsdetail
from .com import COM
from .menge import Menge
from .zeitraum import Zeitraum


@postprocess_docstring
class Ausschreibungslos(COM):
    """
    Eine Komponente zur Abbildung einzelner Lose einer Ausschreibung

    .. raw:: html

        <object data="../_static/images/bo4e/com/Ausschreibungslos.svg" type="image/svg+xml"></object>

    .. HINT::
        `Ausschreibungslos JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Ausschreibungslos.json>`_

    """

    #: Laufende Nummer des Loses
    losnummer: Optional[str] = None
    #: Bezeichnung der Ausschreibung
    bezeichnung: Optional[str] = None
    #: Bezeichnung der Preismodelle in Ausschreibungen für die Energielieferung
    preismodell: Optional[Preismodell] = None

    #: Unterscheidungsmöglichkeiten für die Sparte
    energieart: Optional[Sparte] = None
    #: Aufzählung der Möglichkeiten zur Rechnungslegung in Ausschreibungen
    wunsch_rechnungslegung: Optional[Rechnungslegung] = None
    #: Aufzählung der Möglichkeiten zu Vertragsformen in Ausschreibungen
    wunsch_vertragsform: Optional[Vertragsform] = None
    #: Name des Lizenzpartners
    betreut_durch: Optional[str] = None
    #: Anzahl der Lieferstellen in dieser Ausschreibung
    anzahl_lieferstellen: Optional[int] = None

    #: Die ausgeschriebenen Lieferstellen
    lieferstellen: Optional[list[Ausschreibungsdetail]] = None

    #: Zeitraum, für den die in diesem Los enthaltenen Lieferstellen beliefert werden sollen
    lieferzeitraum: Optional[Zeitraum] = None

    #: Bemerkung des Kunden zum Los
    bemerkung: Optional[str] = None
    #: Gibt den Gesamtjahresverbrauch (z.B. in kWh) aller in diesem Los enthaltenen Lieferstellen an
    gesamt_menge: Optional[Menge] = None
    #: Mindesmenge Toleranzband (kWh, %)
    wunsch_mindestmenge: Optional[Menge] = None
    #: Maximalmenge Toleranzband (kWh, %)
    wunsch_maximalmenge: Optional[Menge] = None

    wiederholungsintervall: Optional[Zeitraum] = None
    """
    In welchem Intervall die Angebotsabgabe wiederholt werden darf.
    Angabe nur gesetzt für die 2. Phase bei öffentlich-rechtlichen Ausschreibungen
    """

    #: Kundenwunsch zur Kündigungsfrist in der Ausschreibung
    wunsch_kuendingungsfrist: Optional[Zeitraum] = None
    #: Kundenwunsch zum Zahlungsziel in der Ausschreibung
    wunsch_zahlungsziel: Optional[Zeitraum] = None



"""
Contains Angebotsvariante and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

import pydantic

from ..enum.angebotsstatus import Angebotsstatus
from ..utils import postprocess_docstring
from .angebotsteil import Angebotsteil
from .betrag import Betrag
from .com import COM
from .menge import Menge


@postprocess_docstring
class Angebotsvariante(COM):
    """
    Führt die verschiedenen Ausprägungen der Angebotsberechnung auf

    .. raw:: html

        <object data="../_static/images/bo4e/com/Angebotsvariante.svg" type="image/svg+xml"></object>

    .. HINT::
        `Angebotsvariante JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Angebotsvariante.json>`_

    """

    #: Gibt den Status eines Angebotes an.
    angebotsstatus: Optional[Angebotsstatus] = None

    #: Datum der Erstellung der Angebotsvariante
    erstellungsdatum: Optional[pydantic.AwareDatetime] = None

    #: Bis zu diesem Zeitpunkt gilt die Angebotsvariante
    bindefrist: Optional[pydantic.AwareDatetime] = None

    teile: Optional[list[Angebotsteil]] = None
    """
    Angebotsteile werden im einfachsten Fall für eine Marktlokation oder Lieferstellenadresse erzeugt.
    Hier werden die Mengen und Gesamtkosten aller Angebotspositionen zusammengefasst.
    Eine Variante besteht mindestens aus einem Angebotsteil.
    """

    #: Aufsummierte Wirkarbeitsmenge aller Angebotsteile
    gesamtmenge: Optional[Menge] = None
    # todo: write a validator for this: https://github.com/Hochfrequenz/BO4E-python/issues/320
    #: Aufsummierte Kosten aller Angebotsteile
    gesamtkosten: Optional[Betrag] = None



"""
Contains Fremdkosten class and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

from pydantic import Field

from ..com.betrag import Betrag
from ..com.fremdkostenblock import Fremdkostenblock
from ..com.zeitraum import Zeitraum
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Fremdkosten(Geschaeftsobjekt):
    """
    Mit diesem BO werden die Fremdkosten, beispielsweise für eine Angebotserstellung oder eine Rechnungsprüfung,
    übertragen.
    Die Fremdkosten enthalten dabei alle Kostenblöcke, die von anderen Marktteilnehmern oder Instanzen erhoben werden.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Fremdkosten.svg" type="image/svg+xml"></object>

    .. HINT::
        `Fremdkosten JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Fremdkosten.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.FREMDKOSTEN
    #: Für diesen Zeitraum wurden die Kosten ermittelt
    gueltigkeit: Optional[Zeitraum] = None
    #: Die Gesamtsumme über alle Kostenblöcke und -positionen
    summe_kosten: Optional[Betrag] = None
    #: In Kostenblöcken werden Kostenpositionen zusammengefasst. Beispiele: Netzkosten, Umlagen, Steuern etc
    kostenbloecke: Optional[list[Fremdkostenblock]] = None



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Messwerterfassung(StrEnum):
    """
    Die Messwerterfassung des Zählers
    """

    FERNAUSLESBAR = "FERNAUSLESBAR"  #: fernauslesbare Zähler
    MANUELL_AUSGELESENE = "MANUELL_AUSGELESENE"  #: manuell ausgelesene Zähler



"""
Contains Sigmoidparameter class and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal
from typing import Optional

from ..utils import postprocess_docstring
from .com import COM

# pylint:disable=invalid-name, too-few-public-methods


@postprocess_docstring
class Sigmoidparameter(COM):
    """
    Die Sigmoid-Funktion, beispielsweise zur Berechnung eines Leistungspreises hat die Form:
    LP=A/(1+(P/B)^C)+D

    .. raw:: html

        <object data="../_static/images/bo4e/com/Sigmoidparameter.svg" type="image/svg+xml"></object>

    .. HINT::
        `Sigmoidparameter JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Sigmoidparameter.json>`_

    """

    A: Optional[Decimal] = None  #: Briefmarke Ortsverteilnetz (EUR/kWh)
    B: Optional[Decimal] = None  #: Wendepunkt für die bepreiste Menge (kW)
    C: Optional[Decimal] = None  #: Exponent (einheitenlos)
    D: Optional[Decimal] = None  #: Briefmarke Transportnetz (EUR/kWh)

    def calculate(self, leistung: Decimal) -> Decimal:
        """
        calculates LP
        :param leistung: Leistung in Kilowatt
        :return: den Sigmoidparameter LP in EUR/kWh
        """
        if self.A is None or self.B is None or self.C is None or self.D is None:
            raise ValueError("Sigmoidparameter is not fully defined")
        return self.A / (1 + (leistung / self.B) ** self.C) + self.D



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class AufAbschlagsziel(StrEnum):
    """
    Der Preis, auf den sich ein Auf- oder Abschlag bezieht.
    """

    ARBEITSPREIS_EINTARIF = "ARBEITSPREIS_EINTARIF"  #: Auf-/Abschlag auf den Arbeitspreis Eintarif
    ARBEITSPREIS_HT = "ARBEITSPREIS_HT"  #: Auf-/Abschlag auf den Arbeitspreis HT
    ARBEITSPREIS_NT = "ARBEITSPREIS_NT"  #: Auf-/Abschlag auf den Arbeitspreis NT
    ARBEITSPREIS_HT_NT = "ARBEITSPREIS_HT_NT"  #: Auf-/Abschlag auf den Arbeitspreis HT und NT
    GRUNDPREIS = "GRUNDPREIS"  #: Auf-/Abschlag auf den Grundpreis
    GESAMTPREIS = "GESAMTPREIS"  #: Auf-/Abschlag auf den Gesamtpreis



"""
utils necessary for reflection/inspection and documentation runs
"""

from typing import TypeVar

from pydantic import BaseModel

from ..version import __gh_version__

T = TypeVar("T", bound=BaseModel)


def postprocess_docstring(cls: type[T]) -> type[T]:
    """
    Postprocess the docstring to inject the __gh_version__ for proper linking of the JSON-schemas.
    Note that doc-strings in Python have to be string literals, so we cannot use f-strings.
    """
    if cls.__doc__ is not None:
        cls.__doc__ = cls.__doc__.format(__gh_version__=__gh_version__)
    return cls



"""
Contains Lastgang class
and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import Field, constr

from ..bo.marktlokation import Marktlokation
from ..bo.messlokation import Messlokation
from ..com.menge import Menge
from ..com.zeitreihenwert import Zeitreihenwert
from ..enum.mengeneinheit import Mengeneinheit
from ..enum.sparte import Sparte
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

# pylint: disable=too-many-instance-attributes, too-few-public-methods


@postprocess_docstring
class Lastgang(Geschaeftsobjekt):
    """
    Modell zur Abbildung eines Lastganges;
    In diesem Modell werden die Messwerte mit einem vollständigen Zeitintervall (zeit_intervall_laenge) angegeben und es bietet daher eine hohe Flexibilität in der Übertragung jeglicher zeitlich veränderlicher Messgrössen.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Lastgang.svg" type="image/svg+xml"></object>

    .. HINT::
        `Lastgang JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Lastgang.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.LASTGANG
    #: Angabe, ob es sich um einen Gas- oder Stromlastgang handelt
    sparte: Optional[Sparte] = None
    #: Definition der gemessenen Größe anhand ihrer Einheit
    messgroesse: Optional[Mengeneinheit] = None
    #:Marktlokation, zu der der Lastgang gehört
    marktlokation: Optional[Marktlokation] = None
    #:Marktlokation, zu der der Lastgang gehört
    messlokation: Optional[Messlokation] = None
    #: Die im Lastgang enthaltenen Messwerte
    werte: Optional[list[Zeitreihenwert]] = None
    #: Versionsnummer des Lastgangs
    version: Optional[str] = None
    #: Die OBIS-Kennzahl für den Wert, die festlegt, welche Größe mit dem Stand gemeldet wird, z.B. '1-0:1.8.1'
    obis_kennzahl: Optional[constr(strict=True)] = None  # type: ignore[valid-type]
    zeit_intervall_laenge: Optional[Menge]



"""
Utility functions for tests.
"""

from pathlib import Path

from pydantic import BaseModel


def parse_file(model: type[BaseModel], path_to_file: Path | str) -> BaseModel:
    """
    Parses the given file as JSON and validates it against the given model.
    If the file is not valid JSON or does not match the model, a ValidationError is raised.
    Returns a validated instance of the given model.
    """
    with open(path_to_file, encoding="utf-8") as file:
        content = file.read()
    return model.model_validate_json(content)



"""
Contains Geschaeftspartner class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-many-instance-attributes, too-few-public-methods, disable=duplicate-code
from typing import Annotated, Optional

from pydantic import Field

from ..com.adresse import Adresse
from ..com.kontaktweg import Kontaktweg
from ..enum.anrede import Anrede
from ..enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from ..enum.organisationstyp import Organisationstyp
from ..enum.titel import Titel
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt
from .person import Person


@postprocess_docstring
class Geschaeftspartner(Geschaeftsobjekt):
    """
    Mit diesem Objekt können Geschäftspartner übertragen werden.
    Sowohl Unternehmen, als auch Privatpersonen können Geschäftspartner sein.
    Hinweis: Marktteilnehmer haben ein eigenes BO, welches sich von diesem BO ableitet.
    Hier sollte daher keine Zuordnung zu Marktrollen erfolgen.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Geschaeftspartner.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geschaeftspartner JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Geschaeftspartner.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.GESCHAEFTSPARTNER
    #: Mögliche Anrede der Person
    anrede: Optional[Anrede] = None
    individuelle_anrede: Optional[str] = None
    """
    Im Falle einer nicht standardisierten Anrede kann hier eine frei definierbare Anrede vorgegeben werden.
    Beispiel: "Vereinsgemeinschaft", "Pfarrer", "Hochwürdigster Herr Abt".
    """
    #: Möglicher Titel der Person
    titel: Optional[Titel] = None
    #: Vorname der Person
    vorname: Optional[str] = None
    #: Nachname (Familienname) der Person
    nachname: Optional[str] = None

    ansprechpartner: Optional[list[Person]] = None
    organisationstyp: Optional[Organisationstyp] = None
    """
    Kennzeichnung ob es sich um ein Gewerbe/Unternehmen, eine Privatperson oder eine andere Art von Organisation handelt.
    """
    organisationsname: Optional[str] = None
    """
    Name der Firma, wenn Gewerbe oder andere Organisation.
    """
    #: Kontaktwege des Geschäftspartners
    kontaktwege: Optional[list[Kontaktweg]] = None
    #: Rollen, die die Geschäftspartner inne haben (z.B. Interessent, Kunde)
    geschaeftspartnerrollen: Optional[list[Geschaeftspartnerrolle]] = None
    #: Handelsregisternummer des Geschäftspartners
    handelsregisternummer: Optional[str] = None
    #: Amtsgericht bzw Handelsregistergericht, das die Handelsregisternummer herausgegeben hat
    amtsgericht: Optional[str] = None
    #: Die Steuer-ID des Geschäftspartners; Beispiel: "DE 813281825"
    umsatzsteuer_id: Optional[str] = None
    #: Die Gläubiger-ID welche im Zahlungsverkehr verwendet wird; Z.B. "DE 47116789"
    glaeubiger_id: Optional[str] = None
    #: Internetseite des Marktpartners
    website: Optional[str] = None
    #: Adresse des Geschäftspartners
    adresse: Optional[Adresse] = None
    #: Todo: Add optional connection to marktteilnehmer as discussed in workshop
    #: not clear what is the best solution here - circular import marktteilnehmer?
    #: discussed in workshop on Feb 6 2024: yes we need the bidirectional option, let's figure out a solution somehow.



from decimal import Decimal

import pytest

from bo4e import (
    Gueltigkeitstyp,
    KriteriumWert,
    Mengeneinheit,
    Preistyp,
    RegionaleGueltigkeit,
    RegionalePreisstaffel,
    RegionaleTarifpreisposition,
    Sigmoidparameter,
    Tarifregionskriterium,
    Waehrungseinheit,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestRegionaleTarifpreisPosition:
    @pytest.mark.parametrize(
        "regionale_tarifpreis_position",
        [
            pytest.param(
                RegionaleTarifpreisposition(
                    preistyp=Preistyp.ARBEITSPREIS_NT,
                    einheit=Waehrungseinheit.EUR,
                    bezugseinheit=Mengeneinheit.KWH,
                    mengeneinheitstaffel=Mengeneinheit.WH,
                    preisstaffeln=[
                        RegionalePreisstaffel(
                            einheitspreis=Decimal(40.0),
                            staffelgrenze_von=Decimal(12.5),
                            staffelgrenze_bis=Decimal(25.0),
                            sigmoidparameter=Sigmoidparameter(),
                            regionale_gueltigkeit=RegionaleGueltigkeit(
                                gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
                                kriteriums_werte=[
                                    KriteriumWert(kriterium=Tarifregionskriterium.POSTLEITZAHL, wert="01069")
                                ],
                            ),
                        ),
                    ],
                ),
                id="all attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, regionale_tarifpreis_position: RegionaleTarifpreisposition) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(regionale_tarifpreis_position)



"""
Contains class Ausschreibungsdetail and corresponding marshmallow schema for de-/serialization
"""

from typing import Optional

from ..enum.zaehlertyp import Zaehlertyp
from ..utils import postprocess_docstring
from .adresse import Adresse
from .com import COM
from .menge import Menge
from .zeitraum import Zeitraum

# pylint: disable=too-few-public-methods, too-many-instance-attributes


@postprocess_docstring
class Ausschreibungsdetail(COM):
    """
    Die Komponente Ausschreibungsdetail wird verwendet um die Informationen zu einer Abnahmestelle innerhalb eines
    Ausschreibungsloses abzubilden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Ausschreibungsdetail.svg" type="image/svg+xml"></object>

    .. HINT::
        `Ausschreibungsdetail JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Ausschreibungsdetail.json>`_

    """

    #: Identifikation einer ausgeschriebenen Marktlokation
    marktlokations_id: Optional[str] = None
    #: In der angegebenen Netzebene wird die Marktlokation versorgt, z.B. MSP für Mittelspannung
    netzebene_lieferung: Optional[str] = None
    #: In der angegebenen Netzebene wird die Lokation gemessen, z.B. NSP für Niederspannung
    netzebene_messung: Optional[str] = None
    #: Die Adresse an der die Marktlokation sich befindet
    marktlokationsadresse: Optional[Adresse] = None
    #: Angefragter Zeitraum für die ausgeschriebene Belieferung
    lieferzeitraum: Optional[Zeitraum] = None

    #: Bezeichnung des zuständigen Netzbetreibers, z.B. 'Stromnetz Hamburg GmbH'
    netzbetreiber: Optional[str] = None
    #: Bezeichnung des Kunden, der die Marktlokation nutzt
    kunde: Optional[str] = None
    #: Die Bezeichnung des Zählers an der Marktlokation
    zaehlernummer: Optional[str] = None
    #: Bezeichnung für die Lokation, z.B. 'Zentraler Einkauf, Hamburg'
    marktlokationsbezeichnung: Optional[str] = None

    #: Spezifikation, um welche Zählertechnik es sich im vorliegenden Fall handelt, z.B. Leistungsmessung
    zaehlertechnik: Optional[Zaehlertyp] = None
    ist_lastgang_vorhanden: Optional[bool] = None
    """
    Zeigt an, ob es zu der Marktlokation einen Lastgang gibt.
    Falls ja, kann dieser abgerufen werden und daraus die Verbrauchswerte ermittelt werden
    """

    #: Prognosewert für die Jahresarbeit der ausgeschriebenen Lokation
    prognose_jahresarbeit: Optional[Menge] = None
    #: Ein Prognosewert für die Arbeit innerhalb des angefragten Lieferzeitraums der ausgeschriebenen Lokation
    prognose_arbeit_lieferzeitraum: Optional[Menge] = None
    #: Prognosewert für die abgenommene maximale Leistung der ausgeschriebenen Lokation
    prognose_leistung: Optional[Menge] = None
    #: Die (evtl. abweichende) Rechnungsadresse
    rechnungsadresse: Optional[Adresse] = None



import pytest

from bo4e import (
    AufAbschlagstyp,
    AufAbschlagsziel,
    Energiemix,
    Preisgarantie,
    RegionalePreisstaffel,
    RegionalerAufAbschlag,
    Tarifeinschraenkung,
    Vertragskonditionen,
    Waehrungseinheit,
    Zeitraum,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestRegionalerAufAbschlag:
    @pytest.mark.parametrize(
        "regionaler_auf_abschlag",
        [
            pytest.param(
                RegionalerAufAbschlag(
                    bezeichnung="Foo",
                    beschreibung="Bar",
                    auf_abschlagstyp=AufAbschlagstyp.RELATIV,
                    auf_abschlagsziel=AufAbschlagsziel.ARBEITSPREIS_HT,
                    einheit=Waehrungseinheit.CT,
                    website="https://www.hochfrequenz.de",
                    zusatzprodukte=["ein standmixer", "ein thermomix"],
                    voraussetzungen=["lecker essen", "mit ökostrom gekocht"],
                    tarifnamensaenderungen="Super-Duper Tarif",
                    staffeln=[RegionalePreisstaffel()],
                    gueltigkeitszeitraum=Zeitraum(),
                    energiemixaenderung=Energiemix(),
                    vertagskonditionsaenderung=Vertragskonditionen(),
                    garantieaenderung=Preisgarantie(),
                    einschraenkungsaenderung=Tarifeinschraenkung(),
                ),
                id="maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, regionaler_auf_abschlag: RegionalerAufAbschlag) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(regionaler_auf_abschlag)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Ausschreibungsstatus(StrEnum):
    """
    Bezeichnungen für die Ausschreibungsphasen
    """

    PHASE1 = "PHASE1"  #: Phase1: Teilnahmewettbewerb
    PHASE2 = "PHASE2"  #: Phase2: Angebotsphase
    PHASE3 = "PHASE3"  #: Phase3: Verhandlungsphase
    PHASE4 = "PHASE4"  #: Phase4: Zuschlagserteilung



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Waermenutzung(StrEnum):
    """
    Wärmenutzung Marktlokation
    """

    SPEICHERHEIZUNG = "SPEICHERHEIZUNG"  #: Speicherheizung
    WAERMEPUMPE = "WAERMEPUMPE"  #: Wärmepumpe
    DIREKTHEIZUNG = "DIREKTHEIZUNG"  #: Direktheizung



import pytest

from bo4e import Region, Regionskriterium
from tests.serialization_helper import assert_serialization_roundtrip


class TestRegion:
    @pytest.mark.parametrize(
        "region",
        [
            pytest.param(
                Region(
                    bezeichnung="Bikini Bottom",
                    positiv_liste=[Regionskriterium()],
                    negativ_liste=[Regionskriterium()],
                ),
                id="max attributes",
            ),
            pytest.param(Region(bezeichnung="Bikini Bottom", positiv_liste=[Regionskriterium()]), id="min attributes"),
        ],
    )
    def test_serialization_roundtrip(self, region: Region) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(region)



from datetime import datetime, timezone

import pytest

from bo4e import Angebotsstatus, Angebotsteil, Angebotsvariante, Betrag, Menge
from tests.serialization_helper import assert_serialization_roundtrip


class TestAngebotsvariante:
    @pytest.mark.parametrize(
        "angebotsvariante",
        [
            pytest.param(
                Angebotsvariante(
                    angebotsstatus=Angebotsstatus.NACHGEFASST,
                    bindefrist=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
                    erstellungsdatum=datetime(2021, 12, 22, 0, 0, 0, tzinfo=timezone.utc),
                    teile=[Angebotsteil()],
                    gesamtmenge=Menge(),
                    gesamtkosten=Betrag(),
                ),
                id="all attributes at first level",  # = min + menge and betrag
            ),
        ],
    )
    def test_serialization_roundtrip(self, angebotsvariante: Angebotsvariante) -> None:
        """
        Test de-/serialisation roundtrip.
        """
        assert_serialization_roundtrip(angebotsvariante)



"""
Contains Preisgarantie class
and corresponding marshmallow schema for de-/serialization
"""

from typing import Optional

from ..enum.preisgarantietyp import Preisgarantietyp
from ..utils import postprocess_docstring
from .com import COM
from .zeitraum import Zeitraum

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Preisgarantie(COM):
    """
    Definition für eine Preisgarantie mit der Möglichkeit verschiedener Ausprägungen.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Preisgarantie.svg" type="image/svg+xml"></object>

    .. HINT::
        `Preisgarantie JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Preisgarantie.json>`_

    """

    #: Festlegung, auf welche Preisbestandteile die Garantie gewährt wird.
    preisgarantietyp: Optional[Preisgarantietyp] = None
    zeitliche_gueltigkeit: Optional[Zeitraum] = None
    """ Zeitraum, bis zu dem die Preisgarantie gilt, z.B. bis zu einem absolutem / fixem Datum
    oder als Laufzeit in Monaten. """

    # optionale attributes
    #: Freitext zur Beschreibung der Preisgarantie.
    beschreibung: Optional[str] = None



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Gebiettyp(StrEnum):
    """
    List of possible Gebiettypen.
    """

    REGELZONE = "REGELZONE"  #: Regelzone
    MARKTGEBIET = "MARKTGEBIET"  #: Marktgebiet
    BILANZIERUNGSGEBIET = "BILANZIERUNGSGEBIET"  #: Bilanzierungsgebiet
    VERTEILNETZ = "VERTEILNETZ"  #: Verteilnetz
    TRANSPORTNETZ = "TRANSPORTNETZ"  #: Transportnetz
    REGIONALNETZ = "REGIONALNETZ"  #: Regionalnetz
    AREALNETZ = "AREALNETZ"  #: Arealnetz
    GRUNDVERSORGUNGSGEBIET = "GRUNDVERSORGUNGSGEBIET"  #: Grundversorgungsgebiet
    VERSORGUNGSGEBIET = "VERSORGUNGSGEBIET"  #: Versorgungsgebiet



"""
Contains functionality to build UML files from code. Also, it contains functions to compile these to svg-files.
Currently, the only supported parser is for Plantuml. The generated `.puml` files can be compiled with kroki by calling
`compile_files_kroki(...)`.
It is designed to work together with `pydantic` and only is tested in this project so far.
"""

import importlib
import inspect
import json
import os
import pkgutil
import re
import shlex
import subprocess
from abc import ABCMeta, abstractmethod
from pathlib import Path
from types import NoneType
from typing import Any, Dict, List, Optional, Tuple, Type, Union, cast, get_args

import networkx as nx
import requests
from annotated_types import Len, MaxLen, MinLen
from pydantic import BaseModel
from pydantic._internal._repr import display_as_type
from pydantic.fields import FieldInfo
from typeguard import TypeCheckError, check_type


# pylint: disable=too-few-public-methods
class _Package:
    """
    Encapsulates scope and color options into a class to make linter and type checker happier
    """

    scope: List[str] | None
    color: str | None

    def __init__(self, *, scope: Optional[List[str]] = None, color: Optional[str] = None):
        self.scope = scope
        self.color = color


pkgs = {
    "bo": _Package(scope=["bo", "com", "enum"], color="#B6D7A8"),
    "com": _Package(scope=["bo", "com", "enum"], color="#E0A86C"),
    "enum": _Package(color="#d1c358"),
}
"""
UML-files will be created only for classes in the packages indicated by the keys of this dict. Additionally for every
package one can define a scope meaning that for every class in e.g. `bo` the UML-file will contain only classes in
packages defined by the value of the dict-entry.
This is for example useful if you want to include bo4e.enum classes in the UMLs but dont want to create UML-files for
them.
"""

pkgs_all = set(sum([pkg.scope if pkg.scope is not None else [] for pkg in pkgs.values()], []))
"""
Contains all packages possibly included inside the network. They are determined by all scopes in `pkgs`.
"""

# accent_color = "#6AA84F"
regex_incl_network = re.compile(r"^bo4e\.(" + "|".join(pkgs_all) + r")")
"""
Regex to include all classes with namespaces matching this pattern. Note that this pattern has to start with `^`.
(because `re.match()` is called) Currently, this pattern matches all classes which appear in the values of `pkgs-scope`.
You can see an example of the regex pattern here: https://regex101.com/r/WibLtS/1
"""

regex_excl_network = re.compile(r"^.*Constrained")
"""
Regex to explicitly exclude all classes with namespaces matching this pattern. Note that this pattern has to start with
`^`. (because `re.match()` is called) Currently, this pattern matches all classes containing `Constrained`.
This is necessary because e.g. ConstrainedStr is an inner class of `BaseModel` (I think - it is from pydantic :))
being inherited by all classes in this project. Therefore, their namespace starts with the respective class e.g.
`bo4e.bo.angebot.Angebot.ConstrainedStr`.
"""

#: Define shorthand for Cardinality type since none of the values have to be provided.
Cardinality = tuple[str, str]

#: Define the link base URI used in svg links
LINK_URI_BASE = "https://bo4e-python.readthedocs.io/en/latest"

# link domain to test links only local.
# LINK_URI_BASE = f"file:///{Path.cwd().parent}/.tox/docs/tmp/html"


class _UMLNetworkABC(nx.MultiDiGraph, metaclass=ABCMeta):
    """
    Defines the abstract base class for all UML-Parsers. Currently, there is only a Plantuml-parser, but you can easily
    add a parser by implementing all abstract methods in a subclass.
    """

    class _DictWrapper(dict[str, Any]):
        """
        This class is needed because `dict` is in cpython. Therefore, it is not possible to simply assign `__hash__`
        function dynamically. Instead, we will construct a python object from `**kwargs`
        """

        def __hash__(self) -> int:  # type:ignore[override]
            """
            This method is designed to be assigned as hash function for dictionaries. In this case specifically used for
            `**kwargs` in `get_node_str` and `get_edge_str` for caching purposes. See those functions for more
            information.
            """
            return hash(json.dumps(self, sort_keys=True))

    def add_class(self, node: str, cls: type[BaseModel]) -> None:
        """
        Adds a class to the UML-Network. It copies the model_fields dictionary because it will possibly be mutated when
        adding superclasses to the network.
        """
        super().add_node(
            node,
            cls=cls,
            fields=(
                {
                    field_name: {"model_field": model_field, "card": None}
                    for field_name, model_field in cls.model_fields.items()
                }
                if hasattr(cls, "model_fields")
                else {}
            ),
        )

    def add_extension(self, node1: str, node2: str) -> None:
        """
        Adds an extension-relation: node1 🠒 node2. All fields of the superclass will be removed in node1 to improve
        clarity.
        """
        super().add_edge(node1, node2, type="extension")
        list(map(self.nodes[node1]["fields"].__delitem__, self.nodes[node2]["fields"].keys()))

    # pylint: disable=too-many-arguments
    def add_association(
        self,
        node1: str,
        node2: str,
        through_field: FieldInfo,
        card1: Optional[Cardinality] = None,
        card2: Optional[Cardinality] = None,
    ) -> None:
        """
        Adds an association-relation. `node1` references `node2` in its field `through_field`. Additionally, you can
        provide information relating to the cardinality of the association.
        """
        super().add_edge(node1, node2, type="association", through_field=through_field, card1=card1, card2=card2)

    def get_node_str(self, node: str, cache: bool = True, **kwargs: Any) -> str:
        """
        Gets the string representation of the node `node`. If `cache` is `true` and this string was already built
        before, the string will be loaded from cache. The cached strings are specific for the set of provided keyword
        arguments. I.e. each set of `kwargs` is expected to result in a different string and therefore cached for each
        different set of `kwargs` separately.
        """
        if cache:
            kwargs = _UMLNetworkABC._DictWrapper(**kwargs)
            if "node_strings" not in self.nodes[node]:
                self.nodes[node]["node_strings"] = {}
            if kwargs not in self.nodes[node]["node_strings"]:
                self.nodes[node]["node_strings"][kwargs] = self._node_to_str(node, **kwargs)
            return self.nodes[node]["node_strings"][kwargs]

        return self._node_to_str(node, **kwargs)

    def get_edge_str(self, node1: str, node2: str, index: int, cache: bool = True, **kwargs: Any) -> str:
        """
        Gets the string representation of the edge. If `cache` is `true` and this string was already built
        before, the string will be loaded from cache. The cached strings are specific for the set of provided keyword
        arguments. I.e. each set of `kwargs` is expected to result in a different string and therefore cached for each
        different set of `kwargs` separately.
        """
        if cache:
            kwargs = _UMLNetworkABC._DictWrapper(**kwargs)
            if "edge_strings" not in self[node1][node2][index]:
                self[node1][node2][index]["edge_strings"] = {}
            if kwargs not in self[node1][node2][index]["edge_strings"]:
                self[node1][node2][index]["edge_strings"][kwargs] = self._edge_to_str(node1, node2, index, **kwargs)
            return self[node1][node2][index]["edge_strings"][kwargs]

        return self._edge_to_str(node1, node2, index, **kwargs)

    @abstractmethod
    def _node_to_str(self, node: str, **kwargs: Any) -> str:
        """
        Returns a string representation of the provided `node`.
        """
        raise NotImplementedError("This method should be overridden.")

    # pylint: disable=too-many-arguments
    @abstractmethod
    def _edge_to_str(self, node1: str, node2: str, index: int, **kwargs: Any) -> str:
        """
        Returns a string representation of the provided edge.
        """
        raise NotImplementedError("This method should be overridden.")

    @abstractmethod
    def network_to_str(self, **kwargs: Any) -> str:
        """
        Returns a string representation of the whole network.
        """
        raise NotImplementedError("This method should be overridden.")

    @abstractmethod
    def get_file_name(self, **kwargs: Any) -> str:
        """
        Returns the desired file-name of this network.
        """
        raise NotImplementedError("This method should be overridden.")

    @staticmethod
    def get_cardinality_string(card: Optional[Cardinality]) -> Optional[str]:
        """
        Parse the cardinality into a readable string e.g. `1..*` or `0..1`
        """
        if card:
            if card[0] == card[1]:
                return f"{card[0]}"
            return f"{card[0]}..{card[1]}"
        return None

    @staticmethod
    def model_field_str(model_field: FieldInfo, card: Optional[Cardinality] = None) -> str:
        """
        Parse the type of the ModelField to a printable string. Copied from
        pydantic._internal._repr.display_as_type
        https://github.com/pydantic/pydantic/blob/58ae1ef77a4bf4276aaa6214aaaaf59455f5e587/pydantic/_internal/_repr.py#L85
        """
        result_str = display_as_type(get_referenced_inner_type(model_field.annotation))
        assert card is not None
        return f"{result_str} [{_UMLNetworkABC.get_cardinality_string(card)}]"

    @staticmethod
    def _remove_last_package_name(namespace: str) -> str:
        """
        E.g. `_remove_last_package_name('bo4e.bo.angebot.Angebot')` -> `bo4e.bo.Angebot`. The only use of this function
        is to make the namespaces in the UML graphs look better (and avoid Plantuml building a subsequent namespace for
        each class - e.g. for `bo4e.bo.angebot.Angebot` would be drawn inside the `bo4e.bo` namespace but in an
        additional namespace `bo4e.bo.angebot`).
        """
        return f'{".".join(namespace.split(".")[0:-2])}.{namespace.split(".")[-1]}'


class PlantUMLNetwork(_UMLNetworkABC):
    """
    UML-Network-Parser inheriting from `_UMLNetworkABC`. This subclass parses the network (i.e. itself) into plantuml
    output.
    """

    def _node_to_str(self, node: str, detailed: bool = True, **kwargs: Any) -> str:
        """
        Parse the class into a string for plantuml output. If `detailed = True` the fields of the class will also be
        included. The class will contain a link to the respective class in the documentation.
        """
        cls_str = (
            f'class "[[{LINK_URI_BASE}/api/bo4e.{node.split(".")[1]}.html#{node}'
            f' {node.split(".")[-1]}]]\\n<size:10>{".".join(node.split(".")[0:-1])}" as {node.split(".")[-1]}'
        )
        if detailed:
            cls_str += " {\n"
            for field_dict in self.nodes[node]["fields"].values():
                model_field: FieldInfo = field_dict["model_field"]
                model_field_inner_type = get_referenced_inner_type(model_field.annotation)
                type_modl_namespace = f"{model_field_inner_type.__module__}.{model_field_inner_type.__name__}"
                if type_modl_namespace in self[node]:
                    # Skip the fields which will appear as references in the graph
                    continue
                type_str = _UMLNetworkABC.model_field_str(model_field, field_dict["card"])
                if model_field.is_required():
                    cls_str += f"\t{model_field.alias} : {type_str}\n"
                else:
                    cls_str += f"\t{model_field.alias} : {type_str} = {model_field.default}\n"
            cls_str += "}"

        return cls_str

    # pylint: disable=too-many-arguments, too-many-locals
    def _edge_to_str(
        self, node1: str, node2: str, index: int, detailed: bool = True, root_node: Optional[str] = None, **kwargs: Any
    ) -> str:
        """
        Parse the connection into a string for plantuml output. If `detailed = True`, cardinalities will also be
        included. The connection's direction will be tweaked with the additional information of the `root_node` of the
        current graph to be drawn.
        """
        # "card" is short for cardinality
        vert_extension = "{node1} -[norank]-|> {node2} #line:gray"
        horz_extension_lr = "{node1} --|> {node2}"
        horz_extension_rl = "{node2} <|-- {node1}"
        vert_association = "{node1} {card1} -[norank]-* {card2} {node2} #line:gray;text:gray : {field}"
        horz_association_lr = "{node1} {card1} --* {card2} {node2} : {field}"
        horz_association_rl = "{node2} {card2} *-- {card1} {node1} : {field}"

        # The only purpose of this is to use less code below by using the get-method with default values
        association: Dict[Optional[str], str] = {
            node1: horz_association_lr,
            node2: horz_association_rl,
        }
        extension: Dict[Optional[str], str] = {
            node1: horz_extension_lr,
            node2: horz_extension_rl,
        }

        node1_str = _UMLNetworkABC._remove_last_package_name(node1)
        node2_str = _UMLNetworkABC._remove_last_package_name(node2)
        if root_node == node1:
            node1_str = "." + node1.split(".")[-1]
        elif root_node == node2:
            node2_str = "." + node2.split(".")[-1]

        if self[node1][node2][index]["type"] == "extension":
            return extension.get(root_node, vert_extension).format(node1=node1_str, node2=node2_str)
        if self[node1][node2][index]["type"] == "association":
            # "card" is short for cardinality
            card1 = None
            card2 = None
            if detailed:
                # ------ Parse the cardinality into readable strings ---------------------------------------------------
                card1 = _UMLNetworkABC.get_cardinality_string(self[node1][node2][index]["card1"])
                card2 = _UMLNetworkABC.get_cardinality_string(self[node1][node2][index]["card2"])
                # ------------------------------------------------------------------------------------------------------
            return association.get(root_node, vert_association).format(
                node1=node1_str,
                node2=node2_str,
                card1=f'"{card1}"' if card1 is not None else "",
                card2=f'"{card2}"' if card2 is not None else "",
                field=self[node1][node2][index]["through_field"].alias,
            )
        raise ValueError(
            f"Illegal edge type for plantuml-parser in ['{node1}']['{node2}']['type']: "
            f"{self[node1][node2][index]['type']}"
        )

    def network_to_str(self, root_node: Optional[str] = None, **kwargs: Any) -> str:
        """
        Build the uml file (content) with class `cls_namespace` treated as root node for this network.
        Classes will be grouped together by namespaces (for instance splitted into `bo` and `com`).
        Only the root node `root_node` will not be included inside its (visual) package namespace.
        """

        regex_pkg = re.compile(r"^bo4e\.(\w+)\.")
        content = "@startuml\nleft to right direction\n\n"
        namespaces: Dict[str, Dict[str, str | bool]] = {}
        # `namespaces` will contain for each package listed in the scope of `pkg` (`pkgs[pkg]["scope"]`) two
        # information:
        # 1. The plantuml-string for the namespaces of the scope containing all respective classes except the root node
        #    (class).
        # 2. A boolean value if there is at least one class in this network in the respective namespace. This is needed
        #    to avoid empty package boxes in the resulting uml-graphs.

        # ------ initialize `namespaces` -------------------------------------------------------------------------------
        for _pkg, pkg_options in pkgs.items():
            namespaces[_pkg] = {
                "str": f'namespace "[[{LINK_URI_BASE}/api/bo4e.{_pkg}.html bo4e.{_pkg}'
                f']]" as bo4e.{_pkg} {pkg_options.color if pkg_options.color is not None else ""} ' + "{\n",
                "empty": True,
            }
        # ------ build the content strings for each node inside this network -------------------------------------------
        for node in self.nodes:
            _pkg = re.match(regex_pkg, node).group(1)  # type:ignore[union-attr]
            if node == root_node:
                content += self.get_node_str(node, detailed=True) + "\n\n"
            else:
                namespaces[_pkg]["str"] += (  # type:ignore[assignment]
                    "\t" + self.get_node_str(node, detailed=False) + "\n"  # type:ignore[operator]
                )
                namespaces[_pkg]["empty"] = False
        # ------ add all non-empty namespace-strings to `content` ------------------------------------------------------
        for namespace in namespaces.values():
            if not namespace["empty"]:
                content += namespace["str"] + "}\n"  # type:ignore[operator]
        content += "\n"
        # ------ add all connections to `content` ----------------------------------------------------------------------
        for edge in self.edges:
            content += self.get_edge_str(edge[0], edge[1], edge[2], detailed=True, root_node=root_node) + "\n"
        # ------ Only show fields on the root node ---------------------------------------------------------------------
        if root_node:
            content += "\nhide members\n" f"show .{root_node.split('.')[-1]} fields\n" "@enduml\n"
        # --------------------------------------------------------------------------------------------------------------
        return content

    def get_file_name(self, root_node: Optional[str] = None, **kwargs: Any) -> str:
        """
        Returns the desired file name for this network with `root_node` treated as the root_node.
        """
        if root_node:
            return root_node.split(".")[-1] + ".puml"
        raise ValueError("You need to provide a root node.")


def write_class_umls(uml_network: _UMLNetworkABC, namespaces_to_parse: List[str], output_dir: Path) -> List[Path]:
    """
    Creates an UML graph for every class listed in `namespaces_to_parse` into `output_dir`.
    For each class a separate uml file will be generated. They include this class and its neighbors.
    Additionally, referenced classes will be included only if `regex_incl_network` matches but explicitly excluded if
    `regex_excl_network` matches the namespace of the respective class (e.g. `bo4e.bo.angebot.Angebot`).
    Only relations between the class and its neighbours will be included.
    Currently, only Plantuml is supported as parser.
    Returns a list of created files.
    """
    path_list: List[Path] = []
    for namespace_to_parse in namespaces_to_parse:
        spl = namespace_to_parse.split(".")
        if pkgs[spl[1]].scope is None:
            raise RuntimeError(f"pkgs[{spl[1]}].scope shouldn't be None")
        file_path = output_dir / "/".join(spl[0:-2])
        file_name = uml_network.get_file_name(root_node=namespace_to_parse)
        uml_subgraph = nx.ego_graph(uml_network, namespace_to_parse, radius=1, undirected=False)
        regex_scope = re.compile(rf'bo4e\.({"|".join(pkgs[spl[1]].scope)})\.')  # type:ignore[arg-type]
        uml_network_scope = cast(
            _UMLNetworkABC,
            # pylint: disable=cell-var-from-loop
            nx.subgraph_view(
                uml_subgraph,
                filter_node=lambda _node: re.match(regex_scope, _node),
                filter_edge=lambda _node1, _node2, _idx: namespace_to_parse in (_node1, _node2),
            ),
        )
        file_content = uml_network_scope.network_to_str(root_node=namespace_to_parse)

        os.makedirs(file_path, exist_ok=True)
        with open(file_path / file_name, "w+", encoding="utf-8") as uml_file:
            uml_file.write(file_content)
            path_list.append(file_path / file_name)

    return path_list


def get_cardinality(model_field: FieldInfo) -> Cardinality:
    """
    Determines the cardinality of a field. This field can either contain a reference to another node in the graph or
    be of another arbitrary type.
    """
    card1: str = "1"
    card2: str = "1"
    try:
        check_type(None, model_field.annotation)
        # The field can be None
        card1 = "0"
    except TypeCheckError:
        pass
    try:
        check_type([], model_field.annotation)
        # The field can be a list
        card1 = "0"
        card2 = "*"
        for metadate in model_field.metadata:
            # min-length and max-length are stored as entries in the metadata list
            if isinstance(metadate, MinLen):
                card1 = str(metadate.min_length)
            elif isinstance(metadate, MaxLen):
                card2 = str(metadate.max_length)
            elif isinstance(metadate, Len):
                card1 = str(metadate.min_length)
                if metadate.max_length is not None:
                    card2 = str(metadate.max_length)
    except TypeCheckError:
        pass
    return card1, card2


def get_referenced_inner_type(type_annotation: Any) -> Any:
    """
    Returns the inner type of a list or optional type annotation.
    Should also handle nested types like Optional[List[...]].
    If the type annotation is not a list or optional type, the type annotation itself is returned.
    """
    while True:
        if not hasattr(type_annotation, "__origin__"):
            return type_annotation
        # pylint: disable=protected-access
        if type_annotation.__origin__ is Union and type_annotation._name == "Optional":
            # Optional is internally a Union with None. We want to ignore the None type.
            generic_alias_args = get_args(type_annotation)
            generic_alias_args = tuple(
                generic_alias_arg for generic_alias_arg in generic_alias_args if generic_alias_arg is not NoneType
            )
        elif type_annotation.__origin__ is list:
            generic_alias_args = get_args(type_annotation)
        elif hasattr(type_annotation, "__metadata__"):
            # This is Annotated[...]
            generic_alias_args = (type_annotation.__origin__,)
        else:
            return type_annotation
        if len(generic_alias_args) > 1:
            raise TypeError(
                f"Could not determine inner type of {type_annotation}: Too many generic alias args {generic_alias_args}"
            )
        if len(generic_alias_args) == 0:
            raise TypeError(
                f"Could not determine inner type of {type_annotation}: "
                f"Undefined generic alias args {generic_alias_args}"
            )
        type_annotation = generic_alias_args[0]


def build_network(module_dir: Path, parser: Type[_UMLNetworkABC]) -> Tuple[_UMLNetworkABC, List[str]]:
    """
    Build a network of the relationships of all classes found in bo4e packages defined by `pkgs` and all classes
    referenced by any class in these packages. Referenced classes will be added only if `regex_incl_network` matches and
    `regex_excl_network` does not match the namespace name of the respective class (e.g. `bo4e.bo.angebot.Angebot`).
    """
    uml_network = parser()
    namespaces_to_parse: List[str] = []
    for pkg, pkg_options in pkgs.items():
        if pkg_options.scope is None:
            continue
        modls = [name for _, name, _ in pkgutil.iter_modules([str(module_dir / pkg)])]
        for modl_name in modls:
            modl_namespace = f"bo4e.{pkg}.{modl_name}"
            modl = importlib.import_module(modl_namespace)
            # pylint: disable=cell-var-from-loop
            cls_list = inspect.getmembers(
                modl, lambda _member: inspect.isclass(_member) and _member.__module__ == modl_namespace
            )
            for name, cls in cls_list:
                modl_namespace = f"{cls.__module__}.{name}"
                namespaces_to_parse.append(modl_namespace)
                if not uml_network.has_node(modl_namespace):
                    _recursive_add_class(cls, modl_namespace, uml_network)
    return uml_network, namespaces_to_parse


def _recursive_add_class(
    cls_cur: type[BaseModel],
    modl_namespace: str,
    uml_network: _UMLNetworkABC,
) -> None:
    """
    Add the specified class `cls_cur` to the `uml_network` and recursively add all classes found in fields and
    bases including all matching `regex_incl_network` but excluding all matching 'regex_excl_network'. If both regex
    are conflicting, the respective class will not be added.
    """
    uml_network.add_class(modl_namespace, cls=cls_cur)
    # ------ add base classes to the network which pass `regex_incl_network` and `regex_excl_network` ------------------
    for parent in cls_cur.__bases__:
        type_modl_namespace = f"{parent.__module__}.{parent.__name__}"
        if re.match(regex_incl_network, type_modl_namespace) and not re.match(regex_excl_network, type_modl_namespace):
            if not uml_network.has_node(type_modl_namespace):
                _recursive_add_class(cast(type[BaseModel], parent), type_modl_namespace, uml_network)
            uml_network.add_extension(
                modl_namespace,
                type_modl_namespace,
            )
    # ------------------------------------------------------------------------------------------------------------------
    # ------ determine references in fields which pass `regex_incl_network` and `regex_excl_network` -------------------
    for field_dict in uml_network.nodes[modl_namespace]["fields"].values():
        model_field: FieldInfo = field_dict["model_field"]
        model_inner_type = get_referenced_inner_type(model_field.annotation)
        # Add cardinality information to the field
        field_card = get_cardinality(model_field)
        field_dict["card"] = field_card
        assert model_field.annotation is not None
        type_modl_namespace = f"{model_inner_type.__module__}.{model_inner_type.__name__}"
        if re.match(regex_incl_network, type_modl_namespace) and not re.match(regex_excl_network, type_modl_namespace):
            if not uml_network.has_node(type_modl_namespace):
                _recursive_add_class(model_inner_type, type_modl_namespace, uml_network)

            uml_network.add_association(
                modl_namespace,
                type_modl_namespace,
                through_field=model_field,
                card1=None,
                card2=field_card,
            )
    # ------------------------------------------------------------------------------------------------------------------


def compile_files_kroki(input_dir: Path, output_dir: Path) -> None:
    """
    Compiles all plantuml files inside `input_dir` (recursive) to svg's in `output_dir` with the same subpath as in
    `input_dir`. Files are compiled using web service of [kroki](https://kroki.io)
    """
    url = "https://kroki.io"
    for root, _, files in os.walk(input_dir):
        for file in files:
            with open(os.path.join(root, file), "r", encoding="utf-8") as uml_file:
                answer = requests.post(
                    url,
                    json={"diagram_source": uml_file.read(), "diagram_type": "plantuml", "output_format": "svg"},
                    timeout=15,
                )
                subdir = root[len(str(input_dir)) + 1 :]
                os.makedirs(output_dir / subdir, exist_ok=True)
                with open(output_dir / subdir / re.sub(r"\.puml$", ".svg", file), "w+", encoding="utf-8") as svg_file:
                    svg_file.write(answer.text)


def compile_files_plantuml(input_dir: Path, output_dir: Path, executable: Path) -> None:
    """
    Compiles all plantuml files inside `input_dir` (not recursive) to svg's in `output_dir`.
    """
    command = f'java -jar "{executable}" "{input_dir}" -svg -o "{output_dir}"'
    subprocess.call(shlex.split(command))



from decimal import Decimal

import pytest

from bo4e import Betrag, Waehrungscode
from tests.serialization_helper import assert_serialization_roundtrip


class TestBetrag:
    @pytest.mark.parametrize(
        "betrag",
        [
            pytest.param(
                Betrag(
                    waehrung=Waehrungscode.EUR,
                    wert=Decimal(12.5),
                ),
            ),
        ],
    )
    def test_regionskriterium_serialization_roundtrip(self, betrag: Betrag) -> None:
        """
        Test de-/serialisation of Regionskriterium with minimal attributes.
        """
        assert_serialization_roundtrip(betrag)



"""
Contains Kosten class and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

from pydantic import Field

from ..com.betrag import Betrag
from ..com.kostenblock import Kostenblock
from ..com.zeitraum import Zeitraum
from ..enum.kostenklasse import Kostenklasse
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt

# pylint: disable=too-many-instance-attributes, too-few-public-methods
# pylint: disable=no-name-in-module


@postprocess_docstring
class Kosten(Geschaeftsobjekt):
    """
    Dieses BO wird zur Übertagung von hierarchischen Kostenstrukturen verwendet.
    Die Kosten werden dabei in Kostenblöcke und diese wiederum in Kostenpositionen strukturiert.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Kosten.svg" type="image/svg+xml"></object>

    .. HINT::
        `Kosten JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Kosten.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.KOSTEN
    #: Klasse der Kosten, beispielsweise Fremdkosten
    kostenklasse: Optional[Kostenklasse] = None
    #: Für diesen Zeitraum wurden die Kosten ermittelt
    gueltigkeit: Optional[Zeitraum] = None
    #: In Kostenblöcken werden Kostenpositionen zusammengefasst. Beispiele: Netzkosten, Umlagen, Steuern etc
    kostenbloecke: Optional[list[Kostenblock]] = None

    #: Die Gesamtsumme über alle Kostenblöcke und -positionen
    summe_kosten: Optional[list[Betrag]] = None



"""
Contains RegionaleTarifpreisposition class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from typing import Optional

from ..enum.mengeneinheit import Mengeneinheit
from ..enum.preistyp import Preistyp
from ..enum.waehrungseinheit import Waehrungseinheit
from ..utils import postprocess_docstring
from .com import COM
from .regionalepreisstaffel import RegionalePreisstaffel


@postprocess_docstring
class RegionaleTarifpreisposition(COM):
    """
    Mit dieser Komponente können Tarifpreise verschiedener Typen im Zusammenhang mit regionalen Gültigkeiten abgebildet
    werden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/RegionaleTarifpreisposition.svg" type="image/svg+xml"></object>

    .. HINT::
        `RegionaleTarifpreisposition JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/RegionaleTarifpreisposition.json>`_

    """

    #: Angabe des Preistypes (z.B. Grundpreis)
    preistyp: Optional[Preistyp] = None
    #: Einheit des Preises (z.B. EURO)
    einheit: Optional[Waehrungseinheit] = None
    #: Größe, auf die sich die Einheit bezieht, beispielsweise kWh, Jahr
    bezugseinheit: Optional[Mengeneinheit] = None
    #: Hier sind die Staffeln mit ihren Preisangaben und regionalen Gültigkeiten definiert
    preisstaffeln: Optional[list[RegionalePreisstaffel]] = None

    #: Gibt an, nach welcher Menge die vorgenannte Einschränkung erfolgt (z.B. Jahresstromverbrauch in kWh)
    mengeneinheitstaffel: Optional[Mengeneinheit] = None



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class ZaehlertypSpezifikation(StrEnum):
    """
    Bei diesem Enum handelt es sich um die Abbildung von besonderen Zählertyp-Spezifikationen der Sparten Strom und Gas.
    """

    EDL40 = "EDL40"  #: EDL40
    EDL21 = "EDL21"  #: EDL21
    SONSTIGER_EHZ = "SONSTIGER_EHZ"  #: SONSTIGER_EHZ
    MME_STANDARD = "MME_STANDARD"  #: MME_STANDARD
    MME_MEDA = "MME_MEDA"  #: MME_MEDA



from datetime import datetime, timezone
from decimal import Decimal

import pytest

from bo4e import Mengeneinheit, Messwertstatus, Verbrauch, Wertermittlungsverfahren
from tests.serialization_helper import assert_serialization_roundtrip


class TestVerbrauch:
    @pytest.mark.parametrize(
        "verbrauch",
        [
            pytest.param(
                Verbrauch(
                    wert=Decimal(40),
                    startdatum=datetime(2021, 12, 1, 0, 0, 0).replace(tzinfo=timezone.utc),
                    enddatum=datetime(2021, 12, 2, 0, 0, 0).replace(tzinfo=timezone.utc),
                    obis_kennzahl="1-0:1.8.1",
                    einheit=Mengeneinheit.KWH,
                    wertermittlungsverfahren=Wertermittlungsverfahren.MESSUNG,
                    messwertstatus=Messwertstatus.ABGELESEN,
                )
            ),
            pytest.param(
                Verbrauch(
                    wert=Decimal(40),
                    obis_kennzahl="1-0:1.8.1",
                    einheit=Mengeneinheit.KWH,
                    wertermittlungsverfahren=Wertermittlungsverfahren.MESSUNG,
                    messwertstatus=Messwertstatus.ABGELESEN,
                ),
            ),
        ],
    )
    def test_serialization_roundtrip(self, verbrauch: Verbrauch) -> None:
        """
        Test de-/serialisation of Verbrauch.
        """
        assert_serialization_roundtrip(verbrauch)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


# pylint: disable=line-too-long
class Geraeteklasse(StrEnum):
    """
    Auflistung möglicher übergreifenden Geräteklassen.
    """

    WANDLER = "WANDLER"  #: Wandler
    KOMMUNIKATIONSEINRICHTUNG = "KOMMUNIKATIONSEINRICHTUNG"  #: Kommunikationseinrichtung
    TECHNISCHE_STEUEREINRICHTUNG = "TECHNISCHE_STEUEREINRICHTUNG"  #: Technische Steuereinrichtung
    MENGENUMWERTER = "MENGENUMWERTER"  #: Mengenumwerter
    SMARTMETER_GATEWAY = "SMARTMETER_GATEWAY"  #: Smartmeter-Gateway
    STEUERBOX = "STEUERBOX"  #: Steuerbox
    ZAEHLEINRICHTUNG = "ZAEHLEINRICHTUNG"  #: Zaehleinrichtung



"""
Contains PreisblattNetnutzung class and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

from pydantic import Field

from ..enum.bilanzierungsmethode import Bilanzierungsmethode
from ..enum.kundengruppe import Kundengruppe
from ..enum.netzebene import Netzebene
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .preisblatt import Preisblatt

# pylint: disable=too-few-public-methods


@postprocess_docstring
class PreisblattNetznutzung(Preisblatt):
    """
    Die Variante des Preisblattmodells zur Abbildung der Netznutzungspreise

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattNetznutzung.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattNetznutzung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/PreisblattNetznutzung.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.PREISBLATTNETZNUTZUNG
    # required attributes (additional to those of Preisblatt)
    #: Die Preise gelten für Marktlokationen der angebebenen Bilanzierungsmethode
    bilanzierungsmethode: Optional[Bilanzierungsmethode] = None
    #: Die Preise gelten für Marktlokationen in der angebebenen Netzebene
    netzebene: Optional[Netzebene] = None
    kundengruppe: Optional[Kundengruppe] = None

    # there are no optional attributes (additionally to those of Preisblatt)



# pylint: disable=missing-module-docstring, line-too-long
from bo4e.enum.strenum import StrEnum


class Voraussetzungen(StrEnum):
    """
    Voraussetzungen, die erfüllt sein müssen, damit dieser Tarif zur Anwendung kommen kann.
    """

    EINZUGSERMAECHTIGUNG = "EINZUGSERMAECHTIGUNG"  #: Einzugsermaechtigung
    ZEITPUNKT = "ZEITPUNKT"  #: Vertrag muss zu einem bestimmten Zeitpunkt noch bestehen
    LIEFERANBINDUNG_EINE = "LIEFERANBINDUNG_EINE"  #: Lieferantenbindung für diese Energieart
    #: Lieferantenbindung für alle Energiearten, die der Versorger anbietet
    LIEFERANBINDUNG_ALLE = "LIEFERANBINDUNG_ALLE"
    GEWERBE = "GEWERBE"  #: Gewerbenachweis
    LASTPROFIL = "LASTPROFIL"  #: Kunde muss einem bestimmten Lastprofil zuzuordnen sein
    ZAEHLERTYP_GROESSE = "ZAEHLERTYP_GROESSE"  #: bestimmter Zaehlertyp oder Zaehlergroeße
    AUSSCHLUSS_GROSSVERBRAUCHER = "AUSSCHLUSS_GROSSVERBRAUCHER"  #: Ausschluss von Großverbrauchern wie Industriekunden oder produzierendes Gewerbe
    NEUKUNDE = "NEUKUNDE"  #: Neukunden ohne bisherige Lieferanbindung
    BESTIMMTE_VERTRAGSFORMALITAETEN = "BESTIMMTE_VERTRAGSFORMALITAETEN"  #: bestimmte Vertragsformalitaeten wie z.B. Anmeldeformular muss an bestimmte Adresse versandt werden
    SELBSTABLESUNG = "SELBSTABLESUNG"  #: Selbstablesung durch den Kunden
    ONLINEVORAUSSETZUNG = "ONLINEVORAUSSETZUNG"  #: Onlinevoraussetzung
    #: Mindestumsatz in einer bestimmten Zeiteinheit wie z.B. Mindestjahresumsatz 2500 EURO
    MINDESTUMSATZ = "MINDESTUMSATZ"
    #: Zusatzprodukt zu bereits bestehendem Lieferverhaeltnis beim Versorger in dieser Energieart
    ZUSATZPRODUKT = "ZUSATZPRODUKT"
    #: geworbener Neukunde muss bestimmte Voraussetzungen erfüllen
    NEUKUNDE_MIT_VORAUSSETZUNGEN = "NEUKUNDE_MIT_VORAUSSETZUNGEN"
    DIREKTVERTRIEB = "DIREKTVERTRIEB"  #: Kunde wird durch Direktvertrieb gewonnen
    ANSCHLUSSART = "ANSCHLUSSART"  #: Anlage mit bestimmter Anschlussart wie z.B. Festanschluss
    ANSCHLUSSWERT = "ANSCHLUSSWERT"  #: bestimmte Anschlusswerte wie z. B. mindestens 30 kW
    #: Alter einer Kundenanlage (z.B. Anlage wurde nach dem 01.01.2005 installiert)
    ALTER_KUNDENANLAGE = "ALTER_KUNDENANLAGE"
    #: bestimmte Anlagebeschaffenheit, wie bivalente Energieversorgung, Geräte o.ä.
    ANLAGEBESCHAFFENHEIT = "ANLAGEBESCHAFFENHEIT"
    #: Betriebsstundenbegrenzung z.B. max 1500h/a oder mindestens 1000h/a
    BETRIEBSSTUNDENBEGRENZUNG = "BETRIEBSSTUNDENBEGRENZUNG"
    FREIGABEZEITEN = "FREIGABEZEITEN"  #: vorgeschriebene Freigabezeiten
    #: Familienstruktur wie z.B. bestimmte Anzahl Kinder oder Personen im Haushalt oder Eheleute
    FAMILIENSTRUKTUR = "FAMILIENSTRUKTUR"
    MITGLIEDSCHAFT = "MITGLIEDSCHAFT"  #: Mitgliedschaft in bestimmten Vereinen oder Verbaenden
    STAATLICHE_FOERDERUNG = "STAATLICHE_FOERDERUNG"  #: staatliche Foerderung wie z.B. Sozialtarif oder Studentenausweis
    #: besondere Verbrauchsstellen wie Kirchen, Vereinsgebaeude usw.
    BESONDERE_VERBRAUCHSSTELLE = "BESONDERE_VERBRAUCHSSTELLE"
    NIEDRIGENERGIE = "NIEDRIGENERGIE"  #: Niedrigenergieaustattung des Hauses
    ORTSTEILE_LIEFERGEBIET = "ORTSTEILE_LIEFERGEBIET"  #: nur für bestimmte Ortsteile in diesem Liefergebiet
    WAERMEBEDARF_ERDGAS = "WAERMEBEDARF_ERDGAS"  #: Wärmebedarf wird nur oder überwiegend mit Erdgas gedeckt
    MAX_ZAEHLER_LIEFERSTELLEN = "MAX_ZAEHLER_LIEFERSTELLEN"  #: beschränkt auf max. Anzahl Zähler oder Abnahmestellen
    #: Lieferungsbeschraenkung auf bestimmte Gasart, wie H-Gas oder L-Gas
    LIEFERUNGSBESCHRAENKUNG_GASART = "LIEFERUNGSBESCHRAENKUNG_GASART"
    #: Kombination von Boni, von denen mindestens einer sehr unwahrscheinlich zu erreichen ist
    KOMBI_BONI = "KOMBI_BONI"
    ALTVERTRAG = "ALTVERTRAG"  #: nur für Altvertraege, die weiterhin gueltig sind
    #: vorgeschriebene Zusatzanlage wie z.B. Solaranlage etc.
    VORGESCHRIEBENE_ZUSATZANLAGE = "VORGESCHRIEBENE_ZUSATZANLAGE"
    MEHRERE_ZAEHLER_ABNAHMESTELLEN = "MEHRERE_ZAEHLER_ABNAHMESTELLEN"  #: mehr als 1 Zaehler oder 1 Abnahmestelle
    #: bestimmter Abnahmefall wie z.B. nur Gemeinschaftsheizungen o.ae.
    BESTIMMTER_ABNAHMEFALL = "BESTIMMTER_ABNAHMEFALL"
    ZUSATZMODALITAET = "ZUSATZMODALITAET"  #: Zahlungsmodalitaet wie z.B. halbjaehrliche Zahlungsweise
    #: Nachweis der Zahlungsfaehigkeit wie z.B. Bonitaetsprüfung
    NACHWEIS_ZAHLUNGSFAEHIGKEIT = "NACHWEIS_ZAHLUNGSFAEHIGKEIT"
    UMSTELLUNG_ENERGIEART = "UMSTELLUNG_ENERGIEART"  #: Umstellung der Energieart



import pytest

from bo4e import (
    Ausschreibungsdetail,
    Ausschreibungslos,
    Menge,
    Preismodell,
    Rechnungslegung,
    Sparte,
    Vertragsform,
    Zeitraum,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestAusschreibungslos:
    @pytest.mark.parametrize(
        "ausschreibungslos",
        [
            pytest.param(
                Ausschreibungslos(
                    losnummer="foo",
                    bezeichnung="bar",
                    bemerkung="asd",
                    preismodell=Preismodell.FESTPREIS,
                    energieart=Sparte.STROM,
                    wunsch_rechnungslegung=Rechnungslegung.MONATSRECHN,
                    wunsch_vertragsform=Vertragsform.DIREKT,
                    betreut_durch="Max Mustermann",
                    anzahl_lieferstellen=17,
                    lieferstellen=[Ausschreibungsdetail()],
                    gesamt_menge=Menge(),
                    wunsch_mindestmenge=Menge(),
                    wunsch_maximalmenge=Menge(),
                    lieferzeitraum=Zeitraum(),
                    wunsch_kuendingungsfrist=Zeitraum(),
                    wunsch_zahlungsziel=Zeitraum(),
                    wiederholungsintervall=Zeitraum(),
                ),
                id="maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, ausschreibungslos: Ausschreibungslos) -> None:
        """
        Test de-/serialisation of Ausschreibungslos
        """
        assert_serialization_roundtrip(ausschreibungslos)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Erzeugungsart(StrEnum):
    """
    Auflistung der Erzeugungsarten von Energie.
    """

    FOSSIL = "FOSSIL"  #: Fossile Brennstoffe
    KWK = "KWK"  #: Kraft-Waerme-Koppelung
    WIND = "WIND"  #: Windkraft
    SOLAR = "SOLAR"  #: Solarenergie
    KERNKRAFT = "KERNKRAFT"  #: Kernkraft
    WASSER = "WASSER"  #: Wasserkraft
    GEOTHERMIE = "GEOTHERMIE"  #: Geothermie
    BIOMASSE = "BIOMASSE"  #: Biomasse
    KOHLE = "KOHLE"  #: Kohle
    GAS = "GAS"  #: Erdgas
    SONSTIGE = "SONSTIGE"  #: Sonstige
    SONSTIGE_EEG = "SONSTIGE_EEG"  #: Sonstige nach EEG
    BIOGAS = "BIOGAS"  #: Biogas
    KLIMANEUTRALES_GAS = "KLIMANEUTRALES_GAS"  #: Klimaneutrales Erdgas



"""
Contains Vertrag class
and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

import pydantic
from pydantic import Field

from ..com.unterschrift import Unterschrift
from ..com.vertragskonditionen import Vertragskonditionen
from ..com.vertragsteil import Vertragsteil
from ..enum.sparte import Sparte
from ..enum.typ import Typ
from ..enum.vertragsart import Vertragsart
from ..enum.vertragsstatus import Vertragsstatus
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt
from .geschaeftspartner import Geschaeftspartner

# pylint: disable=unused-argument
# pylint: disable=no-name-in-module

# pylint: disable=too-many-instance-attributes, too-few-public-methods


@postprocess_docstring
class Vertrag(Geschaeftsobjekt):
    """
    Modell für die Abbildung von Vertragsbeziehungen;
    Das Objekt dient dazu, alle Arten von Verträgen, die in der Energiewirtschaft Verwendung finden, abzubilden.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Vertrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Vertrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Vertrag.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.VERTRAG
    # pylint: disable=duplicate-code
    #: Eine im Verwendungskontext eindeutige Nummer für den Vertrag
    vertragsnummer: Optional[str] = None
    #: Hier ist festgelegt, um welche Art von Vertrag es sich handelt.
    vertragsart: Optional[Vertragsart] = None
    #: Gibt den Status des Vertrags an
    vertragsstatus: Optional[Vertragsstatus] = None
    #: Unterscheidungsmöglichkeiten für die Sparte
    sparte: Optional[Sparte] = None
    #: Gibt an, wann der Vertrag beginnt (inklusiv)
    vertragsbeginn: Optional[pydantic.AwareDatetime] = None
    #: Gibt an, wann der Vertrag (voraussichtlich) endet oder beendet wurde (exklusiv)
    vertragsende: Optional[pydantic.AwareDatetime] = None
    # todo: add von/bis validator
    vertragspartner1: Optional[Geschaeftspartner] = None
    """
    Der "erstgenannte" Vertragspartner.
    In der Regel der Aussteller des Vertrags.
    Beispiel: "Vertrag zwischen Vertragspartner 1 ..."
    """
    vertragspartner2: Optional[Geschaeftspartner] = None
    """
    Der "zweitgenannte" Vertragspartner.
    In der Regel der Empfänger des Vertrags.
    Beispiel "Vertrag zwischen Vertragspartner 1 und Vertragspartner 2".
    """
    vertragsteile: Optional[list[Vertragsteil]] = None
    """
    Der Vertragsteil wird dazu verwendet, eine vertragliche Leistung in Bezug zu einer Lokation
    (Markt- oder Messlokation) festzulegen.
    """

    #: Beschreibung zum Vertrag
    beschreibung: Optional[str] = None
    #: Festlegungen zu Laufzeiten und Kündigungsfristen
    vertragskonditionen: Optional[Vertragskonditionen] = None
    #: Unterzeichner des Vertragspartners 1
    unterzeichnervp1: Optional[list[Unterschrift]] = None
    #: Unterzeichner des Vertragspartners 2
    unterzeichnervp2: Optional[list[Unterschrift]] = None



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Themengebiet(StrEnum):
    """
    Über dieses ENUM kann eine thematische Zuordnung, beispielsweise eines Ansprechpartners, vorgenommen werden.
    """

    ALLGEMEINER_INFORMATIONSAUSTAUSCH = "ALLGEMEINER_INFORMATIONSAUSTAUSCH"  #: Allgemeiner Informationsaustausch
    AN_UND_ABMELDUNG = "AN_UND_ABMELDUNG"  #: An- und Abmeldung
    ANSPRECHPARTNER_ALLGEMEIN = "ANSPRECHPARTNER_ALLGEMEIN"  #: Ansprechpartner Allgemein
    ANSPRECHPARTNER_BDEW_DVGW = "ANSPRECHPARTNER_BDEW_DVGW"  #: Ansprechpartner BDEW/DVGW
    ANSPRECHPARTNER_IT_TECHNIK = "ANSPRECHPARTNER_IT_TECHNIK"  #: Ansprechpartner IT/Technik
    BILANZIERUNG = "BILANZIERUNG"  #: Bilanzierung
    BILANZKREISKOORDINATOR = "BILANZKREISKOORDINATOR"  #: Bilanzkreiskoordinator
    BILANZKREISVERANTWORTLICHER = "BILANZKREISVERANTWORTLICHER"  #: Bilanzkreisverantwortlicher
    #: Datenformate, Zertifikate, Verschlüsselungen
    DATENFORMATE_ZERTIFIKATE_VERSCHLUESSELUNGEN = "DATENFORMATE_ZERTIFIKATE_VERSCHLUESSELUNGEN"
    DEBITORENMANAGEMENT = "DEBITORENMANAGEMENT"  #: Debitorenmanagement
    DEMAND_SIDE_MANAGEMENT = "DEMAND_SIDE_MANAGEMENT"  #: Demand-Side-Management
    EDI_VEREINBARUNG = "EDI_VEREINBARUNG"  #: EDI-Vereinbarung
    EDIFACT = "EDIFACT"  #: EDIFACT
    ENERGIEDATENMANAGEMENT = "ENERGIEDATENMANAGEMENT"  #: Energiedatenmanagement
    FAHRPLANMANAGEMENT = "FAHRPLANMANAGEMENT"  #: Fahrplanmanagement
    ALOCAT = "ALOCAT"  #: Format:ALOCAT
    APERAK = "APERAK"  #: Format:APERAK
    CONTRL = "CONTRL"  #: Format:CONTRL
    INVOIC = "INVOIC"  #: Format:INVOIC
    MSCONS = "MSCONS"  #: Format:MSCONS
    ORDERS = "ORDERS"  #: Format:ORDERS
    ORDERSP = "ORDERSP"  #: Format:ORDERSP
    REMADV = "REMADV"  #: Format:REMADV
    UTILMD = "UTILMD"  #: Format:UTILMD
    GABI = "GABI"  #: GaBi Gas
    GELI = "GELI"  #: GeLi Gas
    GERAETERUECKGABE = "GERAETERUECKGABE"  #: Geräterückgabe
    GERAETEWECHSEL = "GERAETEWECHSEL"  #: Gerätewechsel
    GPKE = "GPKE"  #: GPKE
    INBETRIEBNAHME = "INBETRIEBNAHME"  #: Inbetriebnahme
    KAPAZITAETSMANAGEMENT = "KAPAZITAETSMANAGEMENT"  #: Kapazitätsmanagement
    KLAERFAELLE = "KLAERFAELLE"  #: Klärfälle
    LASTGAENGE_RLM = "LASTGAENGE_RLM"  #: Lastgänge RLM
    LIEFERANTENRAHMENVERTRAG = "LIEFERANTENRAHMENVERTRAG"  #: Lieferantenrahmenvertrag
    LIEFERANTENWECHSEL = "LIEFERANTENWECHSEL"  #: Lieferantenwechsel
    MABIS = "MABIS"  #: MaBiS
    MAHNWESEN = "MAHNWESEN"  #: Mahnwesen
    MARKTGEBIETSVERANTWORTLICHER = "MARKTGEBIETSVERANTWORTLICHER"  #: Marktgebietsverantwortlicher
    MARKTKOMMUNIKATION = "MARKTKOMMUNIKATION"  #: Marktkommunikation
    MEHR_MINDERMENGEN = "MEHR_MINDERMENGEN"  #: Mehr- Mindermengen
    MSB_MDL = "MSB_MDL"  #: MSB - MDL
    NETZABRECHNUNG = "NETZABRECHNUNG"  #: Netzabrechnung
    NETZENTGELTE = "NETZENTGELTE"  #: Netzentgelte
    NETZMANAGEMENT = "NETZMANAGEMENT"  #: Netzmanagement
    RECHT = "RECHT"  #: Recht
    REGULIERUNGSMANAGEMENT = "REGULIERUNGSMANAGEMENT"  #: Regulierungsmanagement
    REKLAMATIONEN = "REKLAMATIONEN"  #: Reklamationen
    SPERREN_ENTSPERREN_INKASSO = "SPERREN_ENTSPERREN_INKASSO"  #: Sperren/Entsperren/Inkasso
    STAMMDATEN = "STAMMDATEN"  #: Stammdaten
    STOERUNGSFAELLE = "STOERUNGSFAELLE"  #: Übermittlung_Störungsfälle
    TECHNISCHE_FRAGEN = "TECHNISCHE_FRAGEN"  #: Technische Fragen
    UMSTELLUNG_INVOIC = "UMSTELLUNG_INVOIC"  #: Umstellung INVOIC
    VERSCHLUESSELUNG_SIGNATUR = "VERSCHLUESSELUNG_SIGNATUR"  #: Verschlüsselung/Signatur
    VERTRAGSMANAGEMENT = "VERTRAGSMANAGEMENT"  #: Vertragsmanagement
    VERTRIEB = "VERTRIEB"  #: Vertrieb
    WIM = "WIM"  #: WiM
    ZAEHLERSTAENDE_SLP = "ZAEHLERSTAENDE_SLP"  #: Zählerstände SLP
    ZAHLUNGSVERKEHR = "ZAHLUNGSVERKEHR"  #: Zahlungsverkehr
    ZUORDNUNGSVEREINBARUNG = "ZUORDNUNGSVEREINBARUNG"  #: Zuordnungsvereinbarung
    EINSPEISUNG = "EINSPEISUNG"  #: Einspeisung
    BEWEGUNGSDATEN = "BEWEGUNGSDATEN"  #: Bewegungsdaten



import pytest

from bo4e import Katasteradresse
from tests.serialization_helper import assert_serialization_roundtrip


class TestKatasteradresse:
    @pytest.mark.parametrize(
        "katasteradresse",
        [
            pytest.param(
                Katasteradresse(gemarkung_flur="hello", flurstueck="world"), id="all attributes at first level"
            ),
        ],
    )
    def test_serialization_roundtrip(self, katasteradresse: Katasteradresse) -> None:
        """
        Test de-/serialisation of Katasteradresse.
        """

        assert_serialization_roundtrip(katasteradresse)



# pylint: disable=missing-module-docstring, line-too-long
from bo4e.enum.strenum import StrEnum


class Kalkulationsmethode(StrEnum):
    """
    Auflistung der verschiedenen Berechnungsmethoden für ein Preisblatt.
    """

    STUFEN = "STUFEN"  #: Stufenmodell, d.h. die Gesamtmenge wird in eine Stufe eingeordnet und für die gesamte Menge gilt der so ermittelte Preis
    ZONEN = "ZONEN"  #: Zonenmodell, d.h. die Gesamtmenge wird auf die Zonen aufgeteilt und für die Teilmengen gilt der jeweilige Preis der Zone.
    VORZONEN_GP = "VORZONEN_GP"  #: Vorzonengrundpreis
    SIGMOID = "SIGMOID"  #: Sigmoidfunktion
    BLINDARBEIT_GT_50_PROZENT = "BLINDARBEIT_GT_50_PROZENT"  #: Blindarbeit oberhalb 50% der Wirkarbeit
    BLINDARBEIT_GT_40_PROZENT = "BLINDARBEIT_GT_40_PROZENT"  #: Blindarbeit oberhalb 40% der Wirkarbeit
    BLINDARBEIT_MIT_FREIMENGE = "BLINDARBEIT_MIT_FREIMENGE"  #: Blindarbeit. Freimenge ist durch cos phi oder prozentualem Anteil der Wirkarbeit angegeben.
    AP_GP_ZONEN = "AP_GP_ZONEN"  #: Arbeits- und Grundpreis gezont
    LP_INSTALL_LEISTUNG = "LP_INSTALL_LEISTUNG"  #: Leistungsentgelt auf Grundlage der installierten Leistung
    AP_TRANSPORT_ODER_VERTEILNETZ = "AP_TRANSPORT_ODER_VERTEILNETZ"  #: AP auf Grundlage Transport- oder Verteilnetz
    AP_TRANSPORT_ODER_VERTEILNETZ_ORTSVERTEILNETZ_SIGMOID = "AP_TRANSPORT_ODER_VERTEILNETZ_ORTSVERTEILNETZ_SIGMOID"  #: AP auf Grundlage Transport- oder Verteilnetz, Ortsverteilnetz über Sigmoid
    LP_JAHRESVERBRAUCH = "LP_JAHRESVERBRAUCH"  #: Leistungsentgelt auf Grundlage des Jahresverbrauchs
    LP_TRANSPORT_ODER_VERTEILNETZ = "LP_TRANSPORT_ODER_VERTEILNETZ"  #: LP auf Grundlage Transport- oder Verteilnetz
    LP_TRANSPORT_ODER_VERTEILNETZ_ORTSVERTEILNETZ_SIGMOID = "LP_TRANSPORT_ODER_VERTEILNETZ_ORTSVERTEILNETZ_SIGMOID"  #: LP auf Grundlage Transport- oder Verteilnetz, Ortsverteilnetz über Sigmoid
    FUNKTIONEN = "FUNKTIONEN"  #: Funktionsbezogene Leistungsermittlung bei Verbräuchen oberhalb der SLP Grenze. (ähnlich Sigmoid)
    VERBRAUCH_UEBER_SLP_GRENZE_FUNKTIONSBEZOGEN_WEITERE_BERECHNUNG_ALS_LGK = "VERBRAUCH_UEBER_SLP_GRENZE_FUNKTIONSBEZOGEN_WEITERE_BERECHNUNG_ALS_LGK"  #: Bei einem Verbrauch über der SLP-Grenze (letzte Staffelgrenze überschritten) erfolgt die Berechnung funktionsbezogen (s.o.) als LGK.



import pytest

from bo4e import (
    Bilanzierungsmethode,
    Dienstleistungstyp,
    Geraet,
    Marktteilnehmer,
    Netzebene,
    PreisblattHardware,
    Preisposition,
    Preisstatus,
    Sparte,
    Zeitraum,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestPreisblattHardware:
    @pytest.mark.parametrize(
        "preisblatt_hardware",
        [
            pytest.param(
                PreisblattHardware(
                    bezeichnung="foo",
                    sparte=Sparte.STROM,
                    preisstatus=Preisstatus.ENDGUELTIG,
                    preispositionen=[Preisposition()],
                    gueltigkeit=Zeitraum(),
                    herausgeber=Marktteilnehmer(),
                    bilanzierungsmethode=Bilanzierungsmethode.TLP_GEMEINSAM,
                    messebene=Netzebene.MSP,
                    inklusive_dienstleistungen=[Dienstleistungstyp.AUSLESUNG_FERNAUSLESUNG_ZUSAETZLICH_MSB],
                    basisgeraet=Geraet(),
                    inklusive_geraete=[Geraet()],
                )
            ),
        ],
    )
    def test_serialization_roundtrip(self, preisblatt_hardware: PreisblattHardware) -> None:
        """
        Test de-/serialisation Preisblatt-Hardware.
        """
        assert_serialization_roundtrip(preisblatt_hardware)



"""
Contains Konzessionsabgabe and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal
from typing import Optional

from ..com.com import COM
from ..enum.abgabeart import AbgabeArt
from ..utils import postprocess_docstring

# pylint: disable=too-few-public-methods, too-many-instance-attributes


@postprocess_docstring
class Konzessionsabgabe(COM):
    """
    Diese Komponente wird zur Übertagung der Details zu einer Konzessionsabgabe verwendet.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Konzessionsabgabe.svg" type="image/svg+xml"></object>

    .. HINT::
        `Konzessionsabgabe JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Konzessionsabgabe.json>`_

    """

    #: Art der Abgabe
    satz: Optional[AbgabeArt] = None

    #: Konzessionsabgabe in E/kWh
    kosten: Optional[Decimal] = None

    #: Gebührenkategorie der Konzessionsabgabe
    kategorie: Optional[str] = None



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e import Geokoordinaten
from tests.serialization_helper import assert_serialization_roundtrip


class TestGeokoordinaten:
    @pytest.mark.parametrize(
        "geokoordinaten",
        [
            pytest.param(
                Geokoordinaten(
                    breitengrad=Decimal(52.52149200439453),
                    laengengrad=Decimal(13.404866218566895),
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_serialization_roundtrip(self, geokoordinaten: Geokoordinaten) -> None:
        assert_serialization_roundtrip(geokoordinaten)

    def test_wrong_datatype(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Geokoordinaten(breitengrad="54,23", laengengrad=-23.2)  # type: ignore[arg-type]

        assert "breitengrad" in str(excinfo.value)



# pylint: disable=missing-module-docstring
from decimal import Decimal
from typing import Annotated, Optional

from humps.main import camelize

# pylint: disable=no-name-in-module
from pydantic import BaseModel, ConfigDict, Field

from bo4e.version import __version__
from bo4e.zusatzattribut import ZusatzAttribut

from ..enum.typ import Typ
from ..utils import postprocess_docstring

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Geschaeftsobjekt(BaseModel):
    """
    Das BO Geschäftsobjekt ist der Master für alle Geschäftsobjekte.
    Alle Attribute, die hier in diesem BO enthalten sind, werden an alle BOs vererbt.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Geschaeftsobjekt.svg" type="image/svg+xml"></object>

    .. HINT::
        `Geschaeftsobjekt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Geschaeftsobjekt.json>`_

    """

    # required attributes
    version: Annotated[
        Optional[str], Field(alias="_version")
    ] = __version__  #: Version der BO-Struktur aka "fachliche Versionierung"
    # src/_bo4e_python_version.py
    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.GESCHAEFTSOBJEKT  #: Der Typ des Geschäftsobjektes
    # bo_typ is used as discriminator f.e. for databases or deserialization

    zusatz_attribute: Optional[list[ZusatzAttribut]] = None
    # zusatz_attribute is a list of ZusatzAttribut objects which are used to store additional information

    # Python internal: The field is not named '_id' because leading underscores are not allowed in pydantic field names.
    # NameError: Fields must not use names with leading underscores; e.g., use 'id' instead of '_id'.
    id: Annotated[Optional[str], Field(alias="_id")] = None
    """
    Eine generische ID, die für eigene Zwecke genutzt werden kann.
    Z.B. könnten hier UUIDs aus einer Datenbank stehen oder URLs zu einem Backend-System.
    """

    #: Hier können IDs anderer Systeme hinterlegt werden (z.B. eine SAP-GP-Nummer oder eine GUID)
    # pylint: disable=duplicate-code
    model_config = ConfigDict(
        alias_generator=camelize,
        populate_by_name=True,
        extra="allow",
        # json_encoders is deprecated, but there is no easy-to-use alternative. The best way would be to create
        # an annotated version of Decimal, but you would have to use it everywhere in the pydantic models.
        # See this issue for more info: https://github.com/pydantic/pydantic/issues/6375
        json_encoders={Decimal: str},
    )
    """
    basic configuration for pydantic's behaviour
    """



from typing import Any

import click
import requests


def ensure_status_code(response: requests.Response) -> None:
    """
    Ensure that the response status code is 2xx
    """
    if not response.ok:
        raise requests.HTTPError(
            f"Request failed with status code {response.status_code}\n"
            f"response-body: {response.json()}\n"
            f"request: {response.request.body}"
        )


def get_release_infos(github_token: str, version: str) -> dict[str, str]:
    """
    Get the release infos from the github api
    """
    response = requests.get(
        f"https://api.github.com/repos/BO4E/BO4E-Python/releases/tags/{version}",
        headers={"Authorization": f"token {github_token}", "Accept": "application/vnd.github.v3+json"},
    )
    ensure_status_code(response)
    return response.json()


def get_latest_release_tag(github_token: str) -> str:
    """
    Get the latest release from the github api
    """
    response = requests.get(
        "https://api.github.com/repos/BO4E/BO4E-Python/releases/latest",
        headers={"Authorization": f"token {github_token}", "Accept": "application/vnd.github.v3+json"},
    )
    ensure_status_code(response)
    return response.json()["tag_name"]


def get_commit_sha(
    github_token: str,
) -> str:
    """
    Get the latest commit sha from the github api
    """
    response = requests.get(
        "https://api.github.com/repos/Hochfrequenz/BO4E-Schemas/git/refs/heads/main",
        headers={"Authorization": f"token {github_token}", "Accept": "application/vnd.github.v3+json"},
    )
    ensure_status_code(response)
    return response.json()["object"]["sha"]


def create_release(
    github_token: str, version: str, commit_sha: str, release_infos: dict[str, Any], is_latest: bool
) -> None:
    """
    Create a release in the github api
    """
    response = requests.post(
        "https://api.github.com/repos/Hochfrequenz/BO4E-Schemas/releases",
        headers={"Authorization": f"token {github_token}", "Accept": "application/vnd.github.v3+json"},
        json={
            "tag_name": version,
            "target_commitish": commit_sha,
            "name": release_infos["name"],
            "body": release_infos["body"],
            "draft": False,
            "prerelease": release_infos["prerelease"],
            "make_latest": str(is_latest),
        },
    )
    ensure_status_code(response)
    return response.json()


@click.command()
@click.option(
    "--github-token",
    "-t",
    help="the github token to use for the release workflow",
    required=True,
    envvar="GITHUB_TOKEN",
    type=click.STRING,
)
@click.option(
    "--version",
    "-v",
    help="the tagged version known inside the release workflow",
    required=True,
    type=click.STRING,
    envvar="VERSION",
)
def main(github_token: str, version: str):
    """create a release in the BO4E-Schemas repository"""
    release_infos = get_release_infos(github_token, version)
    commit_sha = get_commit_sha(github_token)
    is_latest = version == get_latest_release_tag(github_token)
    create_release(github_token, version, commit_sha, release_infos, is_latest)


if __name__ == "__main__":
    main()



"""
Contains PreisblattKonzessionsabgabe class and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

from pydantic import Field

from ..enum.kundengruppeka import KundengruppeKA
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .preisblatt import Preisblatt

# pylint: disable=too-few-public-methods


@postprocess_docstring
class PreisblattKonzessionsabgabe(Preisblatt):
    """
    Die Variante des Preisblattmodells zur Abbildung von allgemeinen Abgaben

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattKonzessionsabgabe.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattKonzessionsabgabe JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/PreisblattKonzessionsabgabe.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.PREISBLATTKONZESSIONSABGABE
    # required attributes (additional to those of Preisblatt)
    #: Kundegruppe anhand derer die Höhe der Konzessionabgabe festgelegt ist
    kundengruppe_k_a: Optional[KundengruppeKA] = None

    # there are no optional attributes (additionally to those of Preisblatt)



"""
Contains Angebotsteil class
and corresponding marshmallow schema for de-/serialization
"""

from typing import Optional

from ..bo.marktlokation import Marktlokation
from ..utils import postprocess_docstring
from .angebotsposition import Angebotsposition
from .betrag import Betrag
from .com import COM
from .menge import Menge
from .zeitraum import Zeitraum

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module


@postprocess_docstring
class Angebotsteil(COM):
    """
    Mit dieser Komponente wird ein Teil einer Angebotsvariante abgebildet.
    Hier werden alle Angebotspositionen aggregiert.
    Angebotsteile werden im einfachsten Fall für eine Marktlokation oder Lieferstellenadresse erzeugt.
    Hier werden die Mengen und Gesamtkosten aller Angebotspositionen zusammengefasst.
    Eine Variante besteht mindestens aus einem Angebotsteil.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Angebotsteil.svg" type="image/svg+xml"></object>

    .. HINT::
        `Angebotsteil JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Angebotsteil.json>`_

    """

    #: Einzelne Positionen, die zu diesem Angebotsteil gehören
    positionen: Optional[list[Angebotsposition]] = None

    #: Identifizierung eines Subkapitels einer Anfrage, beispielsweise das Los einer Ausschreibung
    anfrage_subreferenz: Optional[str] = None
    lieferstellenangebotsteil: Optional[list[Marktlokation]] = None
    """
    Marktlokationen, für die dieses Angebotsteil gilt, falls vorhanden.
    Durch die Marktlokation ist auch die Lieferadresse festgelegt
    """
    #: Summe der Verbräuche aller in diesem Angebotsteil eingeschlossenen Lieferstellen
    gesamtmengeangebotsteil: Optional[Menge] = None
    #: Summe der Jahresenergiekosten aller in diesem Angebotsteil enthaltenen Lieferstellen
    gesamtkostenangebotsteil: Optional[Betrag] = None
    #: Hier kann der Belieferungszeitraum angegeben werden, für den dieser Angebotsteil gilt
    lieferzeitraum: Optional[Zeitraum] = None



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e import Mengeneinheit, Preis, Preisstatus, Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip


class TestPreis:
    def test_wrong_datatype(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Preis(wert="lululululu", einheit=Waehrungseinheit.EUR, bezugswert=Mengeneinheit.KWH)  # type: ignore[arg-type]

        assert "1 validation error" in str(excinfo.value)
        assert "wert" in str(excinfo.value)
        assert "type=decimal_parsing" in str(excinfo.value)

    @pytest.mark.parametrize(
        "preis",
        [
            pytest.param(
                Preis(
                    wert=Decimal(3.50),
                    einheit=Waehrungseinheit.EUR,
                    bezugswert=Mengeneinheit.KWH,
                    status=Preisstatus.ENDGUELTIG,
                ),
            ),
        ],
    )
    def test_serialization_roundtrip(self, preis: Preis) -> None:
        """
        Test de-/serialisation of Preis.
        """
        assert_serialization_roundtrip(preis)



"""
Contains RegionalerAufAbschlag class and corresponding marshmallow schema for de-/serialization
"""

from typing import Optional

from ..enum.aufabschlagstyp import AufAbschlagstyp
from ..enum.aufabschlagsziel import AufAbschlagsziel
from ..enum.waehrungseinheit import Waehrungseinheit

# pylint: disable=R0801
from ..utils import postprocess_docstring
from .com import COM
from .energiemix import Energiemix
from .preisgarantie import Preisgarantie
from .regionalepreisstaffel import RegionalePreisstaffel
from .tarifeinschraenkung import Tarifeinschraenkung
from .vertragskonditionen import Vertragskonditionen
from .zeitraum import Zeitraum

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module


@postprocess_docstring
class RegionalerAufAbschlag(COM):
    """
    Mit dieser Komponente können Auf- und Abschläge verschiedener Typen im Zusammenhang mit regionalen Gültigkeiten
    abgebildet werden.
    Hier sind auch die Auswirkungen auf verschiedene Tarifparameter modelliert, die sich durch die Auswahl eines Auf-
    oder Abschlags ergeben.

    .. raw:: html

        <object data="../_static/images/bo4e/com/RegionalerAufAbschlag.svg" type="image/svg+xml"></object>

    .. HINT::
        `RegionalerAufAbschlag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/RegionalerAufAbschlag.json>`_

    """

    #: Bezeichnung des Auf-/Abschlags
    bezeichnung: Optional[str] = None

    #: Werte für die gestaffelten Auf/Abschläge mit regionaler Eingrenzung
    staffeln: Optional[list[RegionalePreisstaffel]] = None

    #: Beschreibung des Auf-/Abschlags
    beschreibung: Optional[str] = None

    #: Typ des Aufabschlages (z.B. absolut oder prozentual)
    auf_abschlagstyp: Optional[AufAbschlagstyp] = None

    #: Diesem Preis oder den Kosten ist der Auf/Abschlag zugeordnet. Z.B. Arbeitspreis, Gesamtpreis etc.
    auf_abschlagsziel: Optional[AufAbschlagsziel] = None

    #: Gibt an in welcher Währungseinheit der Auf/Abschlag berechnet wird (nur im Falle absoluter Aufschlagstypen).
    einheit: Optional[Waehrungseinheit] = None

    #: Internetseite, auf der die Informationen zum Auf-/Abschlag veröffentlicht sind
    website: Optional[str] = None

    #: Zusatzprodukte, die nur in Kombination mit diesem AufAbschlag erhältlich sind
    zusatzprodukte: Optional[list[str]] = None

    #: Voraussetzungen, die erfüllt sein müssen, damit dieser AufAbschlag zur Anwendung kommen kann
    voraussetzungen: Optional[list[str]] = None

    #: Durch die Anwendung des Auf/Abschlags kann eine Änderung des Tarifnamens auftreten
    tarifnamensaenderungen: Optional[str] = None

    #: Zeitraum, in dem der Abschlag zur Anwendung kommen kann
    gueltigkeitszeitraum: Optional[Zeitraum] = None

    energiemixaenderung: Optional[Energiemix] = None
    """
    Der Energiemix kann sich durch einen AufAbschlag ändern (z.B. zwei Cent Aufschlag für Ökostrom).
    Sollte dies der Fall sein, wird hier die neue Zusammensetzung des Energiemix angegeben.
    """

    vertagskonditionsaenderung: Optional[Vertragskonditionen] = None
    """
    Änderungen in den Vertragskonditionen;
    Falls in dieser Komponenten angegeben, werden die Tarifparameter hiermit überschrieben.
    """

    garantieaenderung: Optional[Preisgarantie] = None
    """
    Änderungen in den Garantievereinbarungen;
    Falls in dieser Komponenten angegeben, werden die Tarifparameter hiermit überschrieben.
    """

    einschraenkungsaenderung: Optional[Tarifeinschraenkung] = None
    """
    Änderungen in den Einschränkungen zum Tarif;
    Falls in dieser Komponenten angegeben, werden die Tarifparameter hiermit überschrieben.
    """



from datetime import datetime, timezone

import pytest

from bo4e import (
    Buendelvertrag,
    Geschaeftspartner,
    Sparte,
    Unterschrift,
    Vertrag,
    Vertragsart,
    Vertragskonditionen,
    Vertragsstatus,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestBuendelvertrag:
    @pytest.mark.parametrize(
        "buendelvertrag",
        [
            pytest.param(
                Buendelvertrag(
                    vertragsnummer="1234567890",
                    vertragsart=Vertragsart.NETZNUTZUNGSVERTRAG,
                    vertragsstatus=Vertragsstatus.AKTIV,
                    sparte=Sparte.STROM,
                    vertragsbeginn=datetime(2021, 4, 30, 13, 45, tzinfo=timezone.utc),
                    vertragsende=datetime(2200, 4, 30, 13, 45, tzinfo=timezone.utc),
                    vertragspartner1=Geschaeftspartner(),
                    vertragspartner2=Geschaeftspartner(),
                    einzelvertraege=[Vertrag()],
                    vertragskonditionen=[Vertragskonditionen()],
                    unterzeichnervp1=[Unterschrift()],
                    unterzeichnervp2=[Unterschrift(), Unterschrift()],
                    beschreibung="Das ist ein Bündelvertrag mit allen optionalen Feldern ausgefüllt.",
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_serialization_roundtrip(self, buendelvertrag: Buendelvertrag) -> None:
        assert_serialization_roundtrip(buendelvertrag)



from decimal import Decimal

import pytest

from bo4e import Steuerbetrag, Steuerkennzeichen, Waehrungscode
from tests.serialization_helper import assert_serialization_roundtrip


class TestSteuerbetrag:
    @pytest.mark.parametrize(
        "steuerbetrag",
        [
            pytest.param(
                Steuerbetrag(
                    steuerkennzeichen=Steuerkennzeichen.UST_7,
                    basiswert=Decimal(100),
                    steuerwert=Decimal(19),
                    waehrung=Waehrungscode.EUR,
                ),
            ),
        ],
    )
    def test_steuerbetrag_required_attributes(self, steuerbetrag: Steuerbetrag) -> None:
        """
        Test de-/serialisation of Steuerbetrag with minimal attributes.
        """
        assert_serialization_roundtrip(steuerbetrag)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Vertragsstatus(StrEnum):
    """
    Abbildung einer Statusinformation für Verträge.
    """

    IN_ARBEIT = "IN_ARBEIT"  #: in Arbeit
    UEBERMITTELT = "UEBERMITTELT"  #: übermittelt
    ANGENOMMEN = "ANGENOMMEN"  #: angenommen
    AKTIV = "AKTIV"  #: aktiv
    ABGELEHNT = "ABGELEHNT"  #: abgelehnt
    WIDERRUFEN = "WIDERRUFEN"  #: widerrufen
    STORNIERT = "STORNIERT"  #: storniert
    GEKUENDIGT = "GEKUENDIGT"  #: gekündigt
    BEENDET = "BEENDET"  #: beendet



"""
Contains AufAbschlag class
and corresponding marshmallow schema for de-/serialization
"""

from typing import Optional

from ..enum.aufabschlagstyp import AufAbschlagstyp
from ..enum.aufabschlagsziel import AufAbschlagsziel
from ..enum.waehrungseinheit import Waehrungseinheit
from ..utils import postprocess_docstring
from .com import COM
from .preisstaffel import Preisstaffel
from .zeitraum import Zeitraum

# pylint: disable=too-few-public-methods, too-many-instance-attributes


@postprocess_docstring
class AufAbschlag(COM):
    """
    Modell für die preiserhöhenden (Aufschlag) bzw. preisvermindernden (Abschlag) Zusatzvereinbarungen,
    die individuell zu einem neuen oder bestehenden Liefervertrag abgeschlossen wurden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/AufAbschlag.svg" type="image/svg+xml"></object>

    .. HINT::
        `AufAbschlag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/AufAbschlag.json>`_

    """

    #: Bezeichnung des Auf-/Abschlags
    bezeichnung: Optional[str] = None
    #: Werte für die gestaffelten Auf/Abschläge.
    staffeln: Optional[list[Preisstaffel]] = None

    #: Beschreibung zum Auf-/Abschlag
    beschreibung: Optional[str] = None
    #: Typ des Aufabschlages (z.B. absolut oder prozentual).
    auf_abschlagstyp: Optional[AufAbschlagstyp] = None
    #: Diesem Preis oder den Kosten ist der Auf/Abschlag zugeordnet. Z.B. Arbeitspreis, Gesamtpreis etc..
    auf_abschlagsziel: Optional[AufAbschlagsziel] = None
    einheit: Optional[Waehrungseinheit] = None
    """ Gibt an in welcher Währungseinheit der Auf/Abschlag berechnet wird. Euro oder Ct..
    (Nur im Falle absoluter Aufschlagstypen). """
    #: Internetseite, auf der die Informationen zum Auf-/Abschlag veröffentlicht sind.
    website: Optional[str] = None
    #: Internetseite, auf der die Informationen zum Auf-/Abschlag veröffentlicht sind.
    gueltigkeitszeitraum: Optional[Zeitraum] = None



import pytest

from bo4e import (
    Adresse,
    Anrede,
    Geschaeftspartner,
    Geschaeftspartnerrolle,
    Kontaktweg,
    Landescode,
    Organisationstyp,
    Person,
    Titel,
)
from tests.serialization_helper import assert_serialization_roundtrip


class TestGeschaeftspartner:
    @pytest.mark.parametrize(
        "geschaeftspartner",
        [
            pytest.param(
                Geschaeftspartner(
                    ansprechpartner=[Person(), Person()],
                    anrede=Anrede.EHELEUTE,
                    individuelle_anrede="Künstler",
                    titel=Titel.PROF_DR,
                    vorname="Hans",
                    nachname="Müller-Schmidt",
                    organisationstyp=Organisationstyp.UNTERNEHMEN,
                    handelsregisternummer="HRB 254466",
                    amtsgericht="Amtsgericht München",
                    kontaktwege=[Kontaktweg()],
                    umsatzsteuer_id="DE267311963",
                    glaeubiger_id="DE98ZZZ09999999999",
                    website="bo4e.de",
                    geschaeftspartnerrollen=[Geschaeftspartnerrolle.DIENSTLEISTER],
                    adresse=Adresse(),
                ),
                id="all attributes at first level",
            ),
            pytest.param(
                Geschaeftspartner(
                    ansprechpartner=[Person(), Person()],
                    anrede=Anrede.EHELEUTE,
                    individuelle_anrede="Künstler",
                    titel=Titel.PROF_DR,
                    vorname="Hans",
                    nachname="Müller-Schmidt",
                    organisationstyp=Organisationstyp.UNTERNEHMEN,
                    handelsregisternummer="HRB 254466",
                    amtsgericht="Amtsgericht Ibiza",
                    kontaktwege=[Kontaktweg()],
                    umsatzsteuer_id="AT12345",
                    geschaeftspartnerrollen=[Geschaeftspartnerrolle.DIENSTLEISTER],
                    adresse=Adresse(
                        postleitzahl="1014", ort="Wien 1", strasse="Ballhausplatz", hausnummer="2", landescode=Landescode.AT  # type: ignore[attr-defined]
                    ),
                ),
                id="Landescode!=DE, DE is default",
            ),
        ],
    )
    def test_serialization_roundtrip(self, geschaeftspartner: Geschaeftspartner) -> None:
        """
        Test de-/serialisation of Geschaeftspartner.
        """

        assert_serialization_roundtrip(geschaeftspartner)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Waehrungseinheit(StrEnum):
    """
    In diesem Enum werden die Währungen und ihre Untereinheiten definiert, beispielsweise für die Verwendung in Preisen.
    """

    EUR = "EUR"  #: Euro
    CT = "CT"  #: Eurocent



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class NetznutzungRechnungstyp(StrEnum):
    """
    Abbildung verschiedener in der INVOIC angegebenen Rechnungstypen.
    """

    ABSCHLUSSRECHNUNG = "ABSCHLUSSRECHNUNG"  #: ABSCHLUSSRECHNUNG
    ABSCHLAGSRECHNUNG = "ABSCHLAGSRECHNUNG"  #: ABSCHLAGSRECHNUNG
    TURNUSRECHNUNG = "TURNUSRECHNUNG"  #: TURNUSRECHNUNG
    MONATSRECHNUNG = "MONATSRECHNUNG"  #: MONATSRECHNUNG
    WIMRECHNUNG = "WIMRECHNUNG"  #: WIMRECHNUNG
    ZWISCHENRECHNUNG = "ZWISCHENRECHNUNG"  #: ZWISCHENRECHNUNG
    INTEGRIERTE_13TE_RECHNUNG = "INTEGRIERTE_13TE_RECHNUNG"  #: INTEGRIERTE_13TE_RECHNUNG
    ZUSAETZLICHE_13TE_RECHNUNG = "ZUSAETZLICHE_13TE_RECHNUNG"  #: ZUSAETZLICHE_13TE_RECHNUNG
    MEHRMINDERMENGENRECHNUNG = "MEHRMINDERMENGENRECHNUNG"  #: MEHRMINDERMENGENRECHNUNG



import pytest

from bo4e import MarktgebietInfo
from tests.serialization_helper import assert_serialization_roundtrip


class TestMarktgebietinfo:
    @pytest.mark.parametrize(
        "marktgebietinfo",
        [
            pytest.param(
                MarktgebietInfo(marktgebiet="Gaspool", marktgebietcode="37Z701133MH0000B"),
            ),
        ],
    )
    def test_serialization(self, marktgebietinfo: MarktgebietInfo) -> None:
        """
        Test de-/serialisation of Marktgebietinfo.
        """

        assert_serialization_roundtrip(marktgebietinfo)



"""
Contains Tarifberechnungsparameter class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal
from typing import Optional

from ..enum.messpreistyp import Messpreistyp
from ..enum.tarifkalkulationsmethode import Tarifkalkulationsmethode
from ..utils import postprocess_docstring
from .com import COM
from .preis import Preis
from .tarifpreis import Tarifpreis

# yes. there is no description in the official docs.
# https://github.com/Hochfrequenz/BO4E-python/issues/328

# pylint: disable=too-few-public-methods, empty-docstring, too-many-instance-attributes


@postprocess_docstring
class Tarifberechnungsparameter(COM):
    """
    In dieser Komponente sind die Berechnungsparameter für die Ermittlung der Tarifkosten zusammengefasst.
    .. raw:: html

        <object data="../_static/images/bo4e/com/Tarifberechnungsparameter.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifberechnungsparameter JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Tarifberechnungsparameter.json>`_

    """

    # there are no required attributes

    #: Gibt an, wie die Einzelpreise des Tarifes zu verarbeiten sind
    berechnungsmethode: Optional[Tarifkalkulationsmethode] = None
    #: True, falls der Messpreis im Grundpreis (GP) enthalten ist
    ist_messpreis_in_grundpreis_enthalten: Optional[bool] = None

    ist_messpreis_zu_beruecksichtigen: Optional[bool] = None
    """
    True, falls bei der Bildung des Durchschnittspreises für die Höchst- und Mindestpreisbetrachtung der Messpreis mit
    berücksichtigt wird
    """

    #: Typ des Messpreises
    messpreistyp: Optional[Messpreistyp] = None

    #: Im Preis bereits eingeschlossene Leistung (für Gas)
    kw_inklusive: Optional[Decimal] = None
    # todo: type decimal is most likely wrong: https://github.com/Hochfrequenz/BO4E-python/issues/327

    #: Intervall, indem die über "kwInklusive" hinaus abgenommene Leistung kostenpflichtig wird (z.B. je 5 kW 20 EURO)
    kw_weitere_mengen: Optional[Decimal] = None
    # todo: type decimal is most likely wrong: https://github.com/Hochfrequenz/BO4E-python/issues/327

    #: Höchstpreis für den Durchschnitts-Arbeitspreis NT
    hoechstpreis_n_t: Optional[Preis] = None
    #: Höchstpreis für den Durchschnitts-Arbeitspreis HT
    hoechstpreis_h_t: Optional[Preis] = None
    #: Mindestpreis für den Durchschnitts-Arbeitspreis
    mindestpreis: Optional[Preis] = None
    #: Liste mit zusätzlichen Preisen, beispielsweise Messpreise und/oder Leistungspreise
    zusatzpreise: Optional[list[Tarifpreis]] = None



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Befestigungsart(StrEnum):
    """
    Befestigungsart von Zählern
    """

    STECKTECHNIK = "STECKTECHNIK"  #: STECKTECHNIK
    DREIPUNKT = "DREIPUNKT"  #: DREIPUNKT
    HUTSCHIENE = "HUTSCHIENE"  #: HUTSCHIENE
    EINSTUTZEN = "EINSTUTZEN"  #: EINSTUTZEN
    ZWEISTUTZEN = "ZWEISTUTZEN"  #: ZWEISTUTZEN



"""
Contains AufAbschlagProOrt class
and corresponding marshmallow schema for de-/serialization
"""

from typing import Optional

from ..utils import postprocess_docstring
from .aufabschlagstaffelproort import AufAbschlagstaffelProOrt
from .com import COM

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module


@postprocess_docstring
class AufAbschlagProOrt(COM):
    """
    Mit dieser Komponente können Auf- und Abschläge verschiedener Typen im Zusammenhang
    mit örtlichen Gültigkeiten abgebildet werden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/AufAbschlagProOrt.svg" type="image/svg+xml"></object>

    .. HINT::
        `AufAbschlagProOrt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/AufAbschlagProOrt.json>`_

    """

    #: Die Postleitzahl des Ortes für den der Aufschlag gilt.
    postleitzahl: Optional[str] = None
    #: Der Ort für den der Aufschlag gilt.
    ort: Optional[str] = None
    #: Die ene't-Netznummer des Netzes in dem der Aufschlag gilt.
    netznr: Optional[str] = None
    #: Werte für die gestaffelten Auf/Abschläge mit regionaler Eingrenzung.
    staffeln: Optional[list[AufAbschlagstaffelProOrt]] = None



import pytest

from bo4e import MarktgebietInfo, StandorteigenschaftenGas
from tests.serialization_helper import assert_serialization_roundtrip

example_standorteigenschaften_gas = StandorteigenschaftenGas(
    netzkontonummern=["GASPOOLNH700xxxx"],
    marktgebiete=[MarktgebietInfo(marktgebiet="Gaspool", marktgebietcode="37Z701133MH0000B")],
)


class TestStandorteigenschaftenGas:
    @pytest.mark.parametrize(
        "standorteigenschaftengas",
        [
            pytest.param(
                StandorteigenschaftenGas(
                    netzkontonummern=["GASPOOLNH700xxxx"],
                    marktgebiete=[MarktgebietInfo(marktgebiet="Gaspool", marktgebietcode="37Z701133MH0000B")],
                ),
            )
        ],
    )
    def test_standorteigenschaftengas_serialization_roundtrip(
        self, standorteigenschaftengas: StandorteigenschaftenGas
    ) -> None:
        assert_serialization_roundtrip(standorteigenschaftengas)



"""
Contains PreisblattDienstleistung class and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

from pydantic import Field

from ..bo.geraet import Geraet
from ..enum.bilanzierungsmethode import Bilanzierungsmethode
from ..enum.dienstleistungstyp import Dienstleistungstyp
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .preisblatt import Preisblatt

# pylint: disable=too-few-public-methods


@postprocess_docstring
class PreisblattDienstleistung(Preisblatt):
    """
    Variante des Preisblattmodells zur Abbildung der Preise für wahlfreie Dienstleistungen

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattDienstleistung.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattDienstleistung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/PreisblattDienstleistung.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.PREISBLATTDIENSTLEISTUNG
    # required attributes (additional to those of Preisblatt)
    #: Die Preise gelten für Marktlokationen der angebebenen Bilanzierungsmethode
    bilanzierungsmethode: Optional[Bilanzierungsmethode] = None
    #: Dienstleistung, für die der Preis abgebildet wird, z.B. Sperrung/Entsperrung
    basisdienstleistung: Optional[Dienstleistungstyp] = None

    #: Hier kann der Preis auf bestimmte Geräte eingegrenzt werden. Z.B. auf die Zählergröße
    geraetedetails: Optional[Geraet] = None

    #: Weitere Dienstleistungen, die im Preis enthalten sind
    inklusive_dienstleistungen: Optional[list[Dienstleistungstyp]] = None



"""
Contains PreisblattMessung class and corresponding marshmallow schema for de-/serialization
"""

from typing import Annotated, Optional

from pydantic import Field

from ..bo.geraet import Geraet
from ..bo.zaehler import Zaehler
from ..enum.bilanzierungsmethode import Bilanzierungsmethode
from ..enum.dienstleistungstyp import Dienstleistungstyp
from ..enum.netzebene import Netzebene
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .preisblatt import Preisblatt

# pylint: disable=too-few-public-methods


@postprocess_docstring
class PreisblattMessung(Preisblatt):
    """
    Variante des Preisblattmodells zur Abbildung der Preise des Messstellenbetriebs und damit verbundener Leistungen

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattMessung.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattMessung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/PreisblattMessung.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.PREISBLATTMESSUNG
    # required attributes (additional to those of Preisblatt)
    #: Die Preise gelten für Marktlokationen der angebebenen Bilanzierungsmethode
    bilanzierungsmethode: Optional[Bilanzierungsmethode] = None
    #: Die Preise gelten für Messlokationen in der angebebenen Netzebene
    messebene: Optional[Netzebene] = None

    #: Der Preis betrifft den hier angegebenen Zähler, z.B. einen Drehstromzähler
    zaehler: Optional[Zaehler] = None

    #: Im Preis sind die hier angegebenen Dienstleistungen enthalten, z.B. Jährliche Ablesung
    inklusive_dienstleistungen: Optional[list[Dienstleistungstyp]] = None

    #: Im Preis sind die hier angegebenen Geräte mit enthalten, z.B. ein Wandler
    inklusive_geraete: Optional[list[Geraet]] = None



"""
Contains Steuerbetrag class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal
from typing import Optional

from ..enum.steuerkennzeichen import Steuerkennzeichen
from ..enum.waehrungscode import Waehrungscode
from ..utils import postprocess_docstring
from .com import COM

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Steuerbetrag(COM):
    """
    Abbildung eines Steuerbetrages.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Steuerbetrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Steuerbetrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Steuerbetrag.json>`_

    """

    #: Kennzeichnung des Steuersatzes, bzw. Verfahrens.
    steuerkennzeichen: Optional[Steuerkennzeichen] = None
    #: Nettobetrag für den die Steuer berechnet wurde. Z.B. 100
    basiswert: Optional[Decimal] = None
    #: Aus dem Basiswert berechnete Steuer. Z.B. 19 (bei UST_19)
    steuerwert: Optional[Decimal] = None
    #: Währung. Z.B. Euro.
    waehrung: Optional[Waehrungscode] = None



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Messwertstatuszusatz(StrEnum):
    """
    Aufzählung von zusätzlichen Informationen zum Status, beispielsweise in Lastgängen oder Zählwerkständen.
    """

    Z84_LEERSTAND = "Z84_LEERSTAND"  #: Z84_LEERSTAND
    Z85_REALERZAEHLERUEBERLAUFGEPRUEFT = "Z85_REALERZAEHLERUEBERLAUFGEPRUEFT"  #: Z85_REALERZAEHLERUEBERLAUFGEPRUEFT
    Z86_PLAUSIBELWGKONTROLLABLESUNG = "Z86_PLAUSIBELWGKONTROLLABLESUNG"  #: Z86_PLAUSIBELWGKONTROLLABLESUNG
    Z87_PLAUSIBELWGKUNDENHINWEIS = "Z87_PLAUSIBELWGKUNDENHINWEIS"  #: Z87_PLAUSIBELWGKUNDENHINWEIS
    ZC3_AUSTAUSCHDESERSATZWERTES = "ZC3_AUSTAUSCHDESERSATZWERTES"  #: ZC3_AUSTAUSCHDESERSATZWERTES
    Z88_VERGLEICHSMESSUNG_GEEICHT = "Z88_VERGLEICHSMESSUNG(GEEICHT)"  #: Z88_VERGLEICHSMESSUNG(GEEICHT)
    Z89_VERGLEICHSMESSUNG_NICHTGEEICHT = "Z89_VERGLEICHSMESSUNG(NICHTGEEICHT)"  #: Z89_VERGLEICHSMESSUNG(NICHTGEEICHT)
    #: Z90_MESSWERTNACHBILDUNGAUSGEEICHTENWERTEN
    Z90_MESSWERTNACHBILDUNGAUSGEEICHTENWERTEN = "Z90_MESSWERTNACHBILDUNGAUSGEEICHTENWERTEN"
    #: Z91_MESSWERTNACHBILDUNGAUSNICHTGEEICHTENWERTEN
    Z91_MESSWERTNACHBILDUNGAUSNICHTGEEICHTENWERTEN = "Z91_MESSWERTNACHBILDUNGAUSNICHTGEEICHTENWERTEN"
    Z92_INTERPOLATION = "Z92_INTERPOLATION"  #: Z92_INTERPOLATION
    Z93_HALTEWERT = "Z93_HALTEWERT"  #: Z93_HALTEWERT
    Z94_BILANZIERUNGNETZABSCHNITT = "Z94_BILANZIERUNGNETZABSCHNITT"  #: Z94_BILANZIERUNGNETZABSCHNITT
    Z95_HISTORISCHEMESSWERTE = "Z95_HISTORISCHEMESSWERTE"  #: Z95_HISTORISCHEMESSWERTE
    ZJ2_STATISTISCHEMETHODE = "ZJ2_STATISTISCHEMETHODE"  #: ZJ2_STATISTISCHEMETHODE
    Z74_KEINZUGANG = "Z74_KEINZUGANG"  #: Z74_KEINZUGANG
    Z75_KOMMUNIKATIONSSTOERUNG = "Z75_KOMMUNIKATIONSSTOERUNG"  #: Z75_KOMMUNIKATIONSSTOERUNG
    Z76_NETZAUSFALL = "Z76_NETZAUSFALL"  #: Z76_NETZAUSFALL
    Z77_SPANNUNGSAUSFALL = "Z77_SPANNUNGSAUSFALL"  #: Z77_SPANNUNGSAUSFALL
    Z78_GERAETEWECHSEL = "Z78_GERAETEWECHSEL"  #: Z78_GERAETEWECHSEL
    Z79_KALIBRIERUNG = "Z79_KALIBRIERUNG"  #: Z79_KALIBRIERUNG
    #: Z80_GERAETARBEITETAUSSERHALBDERBETRIEBSBEDINGUNGEN
    Z80_GERAETARBEITETAUSSERHALBDERBETRIEBSBEDINGUNGEN = "Z80_GERAETARBEITETAUSSERHALBDERBETRIEBSBEDINGUNGEN"
    Z81_MESSEINRICHTUNGGESTOERT_DEFEKT = "Z81_MESSEINRICHTUNGGESTOERT_DEFEKT"  #: Z81_MESSEINRICHTUNGGESTOERT_DEFEKT
    Z82_UNSICHERHEITMESSUNG = "Z82_UNSICHERHEITMESSUNG"  #: Z82_UNSICHERHEITMESSUNG
    #: Z98_BERUECKSICHTIGUNGSTOERMENGENZAEHLWERK
    Z98_BERUECKSICHTIGUNGSTOERMENGENZAEHLWERK = "Z98_BERUECKSICHTIGUNGSTOERMENGENZAEHLWERK"
    Z99_MENGENUMWERTUNGUNVOLLSTAENDIG = "Z99_MENGENUMWERTUNGUNVOLLSTAENDIG"  #: Z99_MENGENUMWERTUNGUNVOLLSTAENDIG
    ZA0_UHRZEITGESTELLT_SYNCHRONISATION = "ZA0_UHRZEITGESTELLT_SYNCHRONISATION"  #: ZA0_UHRZEITGESTELLT_SYNCHRONISATION
    ZA1_MESSWERTUNPLAUSIBEL = "ZA1_MESSWERTUNPLAUSIBEL"  #: ZA1_MESSWERTUNPLAUSIBEL
    ZC2_TARIFSCHALTGERAETDEFEKT = "ZC2_TARIFSCHALTGERAETDEFEKT"  #: ZC2_TARIFSCHALTGERAETDEFEKT
    #: ZC4_IMPULSWERTIGKEITNICHTAUSREICHEND
    ZC4_IMPULSWERTIGKEITNICHTAUSREICHEND = "ZC4_IMPULSWERTIGKEITNICHTAUSREICHEND"
    ZA3_FALSCHERWANDLERFAKTOR = "ZA3_FALSCHERWANDLERFAKTOR"  #: ZA3_FALSCHERWANDLERFAKTOR
    ZA4_FEHLERHAFTEABLESUNG = "ZA4_FEHLERHAFTEABLESUNG"  #: ZA4_FEHLERHAFTEABLESUNG
    ZA5_AENDERUNGDERBERECHNUNG = "ZA5_AENDERUNGDERBERECHNUNG"  #: ZA5_AENDERUNGDERBERECHNUNG
    ZA6_UMBAUDERMESSLOKATION = "ZA6_UMBAUDERMESSLOKATION"  #: ZA6_UMBAUDERMESSLOKATION
    ZA7_DATENBEARBEITUNGSFEHLER = "ZA7_DATENBEARBEITUNGSFEHLER"  #: ZA7_DATENBEARBEITUNGSFEHLER
    ZA8_BRENNWERTKORREKTUR = "ZA8_BRENNWERTKORREKTUR"  #: ZA8_BRENNWERTKORREKTUR
    ZA9_Z_ZAHL_KORREKTUR = "ZA9_Z-ZAHL-KORREKTUR"  #: ZA9_Z-ZAHL-KORREKTUR
    ZB0_STOERUNG_DEFEKTMESSEINRICHTUNG = "ZB0_STOERUNG_DEFEKTMESSEINRICHTUNG"  #: ZB0_STOERUNG_DEFEKTMESSEINRICHTUNG
    ZB9_AENDERUNGTARIFSCHALTZEITEN = "ZB9_AENDERUNGTARIFSCHALTZEITEN"  #: ZB9_AENDERUNGTARIFSCHALTZEITEN
    ZG3_UMSTELLUNGGASQUALITAET = "ZG3_UMSTELLUNGGASQUALITAET"  #: ZG3_UMSTELLUNGGASQUALITAET



import pytest

from bo4e import TarifpreispositionProOrt, TarifpreisstaffelProOrt
from tests.serialization_helper import assert_serialization_roundtrip


class TestTarifpreispositionProOrt:
    @pytest.mark.parametrize(
        "tarifpreispositionproort",
        [
            pytest.param(
                TarifpreispositionProOrt(
                    postleitzahl="82031",
                    ort="Grünwald",
                    netznr="0815",
                    preisstaffeln=[TarifpreisstaffelProOrt()],
                ),
                id="minimal and maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifpreispositionproort: TarifpreispositionProOrt) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarifpreispositionproort)



"""
Contains Ausschreibung class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import Annotated, Optional

import pydantic
from pydantic import Field

from ..com.ausschreibungslos import Ausschreibungslos
from ..com.zeitraum import Zeitraum
from ..enum.ausschreibungsportal import Ausschreibungsportal
from ..enum.ausschreibungsstatus import Ausschreibungsstatus
from ..enum.ausschreibungstyp import Ausschreibungstyp
from ..enum.typ import Typ
from ..utils import postprocess_docstring
from .geschaeftsobjekt import Geschaeftsobjekt
from .geschaeftspartner import Geschaeftspartner


@postprocess_docstring
class Ausschreibung(Geschaeftsobjekt):
    """
    Das BO Ausschreibung dient zur detaillierten Darstellung von ausgeschriebenen Energiemengen in der Energiewirtschaft

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Ausschreibung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Ausschreibung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/bo/Ausschreibung.json>`_

    """

    typ: Annotated[Optional[Typ], Field(alias="_typ")] = Typ.AUSSCHREIBUNG
    #: Vom Herausgeber der Ausschreibung vergebene eindeutige Nummer
    ausschreibungsnummer: Optional[str] = None
    #: Aufzählung für die Typisierung von Ausschreibungen
    ausschreibungstyp: Optional[Ausschreibungstyp] = None
    #: Bezeichnungen für die Ausschreibungsphasen
    ausschreibungsstatus: Optional[Ausschreibungsstatus] = None
    #: Kennzeichen, ob die Ausschreibung kostenpflichtig ist
    ist_kostenpflichtig: Optional[bool] = None
    #: Gibt den Veröffentlichungszeitpunkt der Ausschreibung an
    veroeffentlichungszeitpunkt: Optional[pydantic.AwareDatetime] = None
    ausschreibender: Optional[Geschaeftspartner] = None
    """
    Mit diesem Objekt können Geschäftspartner übertragen werden.
    Sowohl Unternehmen, als auch Privatpersonen können Geschäftspartner sein
    """
    abgabefrist: Optional[Zeitraum] = None
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher entweder eine Dauer oder ein Zeitraum in Form von Start und Ende angegeben sein
    """
    bindefrist: Optional[Zeitraum] = None
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher entweder eine Dauer oder ein Zeitraum in Form von Start und Ende angegeben sein
    """
    #: Die einzelnen Lose, aus denen sich die Ausschreibung zusammensetzt
    lose: Optional[list[Ausschreibungslos]] = None

    #: Aufzählung der unterstützten Ausschreibungsportale
    ausschreibungportal: Optional[Ausschreibungsportal] = None
    #: Internetseite, auf der die Ausschreibung veröffentlicht wurde (falls vorhanden)
    webseite: Optional[str] = None



from datetime import datetime, timezone
from decimal import Decimal

import pytest

from bo4e import Betrag, Kostenposition, Menge, Mengeneinheit, Preis, Preisstatus, Waehrungscode, Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip


class TestKostenposition:
    @pytest.mark.parametrize(
        "kostenposition",
        [
            pytest.param(
                Kostenposition(
                    positionstitel="Vadders Kostenposition",
                    von=datetime(2013, 5, 1, tzinfo=timezone.utc),
                    bis=datetime(2014, 5, 1, tzinfo=timezone.utc),
                    artikelbezeichnung="Deim Vadder sei Kostenposition",
                    zeitmenge=Menge(),
                    einzelpreis=Preis(
                        wert=Decimal(3.50),
                        einheit=Waehrungseinheit.EUR,
                        bezugswert=Mengeneinheit.KWH,
                        status=Preisstatus.ENDGUELTIG,
                    ),
                    betrag_kostenposition=Betrag(
                        waehrung=Waehrungscode.EUR,
                        wert=Decimal(12.5),
                    ),
                    menge=Menge(),
                    artikeldetail="foo",
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_serialization_roundtrip(self, kostenposition: Kostenposition) -> None:
        """
        Test de-/serialisation of Kostenposition.
        """
        assert_serialization_roundtrip(kostenposition)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Zaehlertyp(StrEnum):
    """
    Bei diesem Enum handelt es sich um die Abbildung von Zählertypen der Sparten Strom und Gas.
    """

    #: Drehstromzähler - dieser Zählertyp wird für dreiphasige Ferraris-Zähler verwendet
    DREHSTROMZAEHLER = "DREHSTROMZAEHLER"
    BALGENGASZAEHLER = "BALGENGASZAEHLER"  #: Balgengaszähler
    DREHKOLBENZAEHLER = "DREHKOLBENZAEHLER"  #: Drehkolbengaszähler
    LEISTUNGSZAEHLER = "LEISTUNGSZAEHLER"  #: leistungsmessender Zähler
    MAXIMUMZAEHLER = "MAXIMUMZAEHLER"  #: Maximumzähler
    TURBINENRADGASZAEHLER = "TURBINENRADGASZAEHLER"  #: Turbinenradgaszähler
    ULTRASCHALLGASZAEHLER = "ULTRASCHALLGASZAEHLER"  #: Ultraschallgaszähler
    #: Wechselstromzähler - dieser Zählertyp wird für einphasige Ferraris-Zähler verwendet
    WECHSELSTROMZAEHLER = "WECHSELSTROMZAEHLER"
    MODERNE_MESSEINRICHTUNG = "MODERNE_MESSEINRICHTUNG"  #: Moderne Messeinrichtung
    INTELLIGENTES_MESSSYSTEM = "INTELLIGENTES_MESSSYSTEM"  #: Intelligentes Messsystem
    ELEKTRONISCHER_ZAEHLER = "ELEKTRONISCHER_ZAEHLER"  #: Elektronischer Zähler
    WIRBELGASZAEHLER = "WIRBELGASZAEHLER"  #: Wirbelgaszähler



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class ArithmetischeOperation(StrEnum):
    """
    Mit dieser Aufzählung können arithmetische Operationen festgelegt werden.
    """

    ADDITION = "ADDITION"  #: Es wird addiert
    SUBTRAKTION = "SUBTRAKTION"  #: Es wird subtrahiert
    MULTIPLIKATION = "MULTIPLIKATION"  #: Es wird multipliziert
    DIVISION = "DIVISION"  #: Es wird dividiert



from datetime import datetime, timezone

import pytest

from bo4e import Betrag, Fremdkostenposition, Menge, Preis
from tests.serialization_helper import assert_serialization_roundtrip


class TestFremdkostenposition:
    @pytest.mark.parametrize(
        "fremdkostenposition",
        [
            pytest.param(
                Fremdkostenposition(
                    positionstitel="Vadders Preisstaffel",
                    von=datetime(2013, 5, 1, tzinfo=timezone.utc),
                    bis=datetime(2014, 5, 1, tzinfo=timezone.utc),
                    artikelbezeichnung="Deim Vadder sei Preisstaffel",
                    link_preisblatt="http://foo.bar/",
                    zeitmenge=Menge(),
                    einzelpreis=Preis(),
                    betrag_kostenposition=Betrag(),
                    menge=Menge(),
                    artikeldetail="foo",
                    marktpartnercode="986543210123",
                    marktpartnername="Mein MP",
                    gebietcode_eic="not an eic code but validation will follow in ticket 146",
                ),
                id="all attributes at first level",
            ),
        ],
    )
    def test_serialization_roundtrip(self, fremdkostenposition: Fremdkostenposition) -> None:
        """
        Test de-/serialisation of Fremdkostenposition.
        """
        assert_serialization_roundtrip(fremdkostenposition)
